def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import PX, P, X

    for a in [2, 3]:
        m.utility_co[a] = +P("hhinc#2,3") * X("hhinc")
    for a in [4, 5, 6]:
        m.utility_co[a] = +P(f"hhinc#{a}") * X("hhinc")

    d.dc.alts_mapping

    for a, name in d.dc.alts_mapping.items():
        if a == 1:
            continue
        m.utility_co[a] += (
            +P("ASC_" + name)
            + P("vehbywrk_" + name) * X("vehbywrk")
            + P("wkcbd_" + name) * X("wkccbd + wknccbd")
        )

    m.utility_ca = (
        +PX("totcost")
        + P("motorized_time") * X("(altid <= 4) * tottime")
        + P("nonmotorized_time") * X("(altid > 4) * tottime")
        + P("motorized_ovtbydist") * X("(altid <= 4) * ovtt/dist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 13, CBD"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, method="bhhh")
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        ("Ownership", "vehbywrk.*"),
        ("Zonal", "wkcbd.*"),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_Bike": -1.65086990936308,
        "ASC_SR2": -1.6343459309594404,
        "ASC_SR3+": -3.536895423106506,
        "ASC_Transit": -0.20182268467513967,
        "ASC_Walk": 0.08378500087016123,
        "hhinc#2,3": -0.0017353041895238406,
        "hhinc#4": -0.006149027560100709,
        "hhinc#5": -0.011116299683489442,
        "hhinc#6": -0.007835540009972525,
        "motorized_ovtbydist": -0.15011160599438625,
        "motorized_time": -0.028598343252833735,
        "nonmotorized_time": -0.04641360594486308,
        "totcost": -0.0032855739180273234,
        "vehbywrk_Bike": -0.6979736536562279,
        "vehbywrk_SR2": -0.4154516212033234,
        "vehbywrk_SR3+": -0.21205944593613626,
        "vehbywrk_Transit": -0.9109101165995294,
        "vehbywrk_Walk": -0.7194478370164099,
        "wkcbd_Bike": 0.3759603983833265,
        "wkcbd_SR2": 0.2558900818841308,
        "wkcbd_SR3+": 1.0573013291451367,
        "wkcbd_Transit": 1.3562157325427109,
        "wkcbd_Walk": 0.17467164791819956,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(expected_x[k], 2e-2), (
            f"{k}, {revealed_x[k] / expected_x[k]}"
        )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations


class SimpleAttribute:
    """
    A simple attribute descriptor.

    Use this class to create a simple attribute descriptor that circumvents
    a block on arbitrary attribute assignment.
    """

    def __init__(self, default: type = None):
        self.default = default

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            try:
                result = getattr(instance, self.private_name)
            except AttributeError:
                if self.default is None:
                    result = None
                else:
                    setattr(instance, self.private_name, self.default())
                    result = getattr(instance, self.private_name)
            return result

    def __set__(self, instance, values):
        if instance is not None:
            if self.default is not None and not isinstance(values, self.default):
                raise TypeError(
                    f"when assigning to {instance.__class__.__name__}.{self.name}, "
                    f"expected {self.default.__name__}, got {type(values).__name__}"
                )
            setattr(instance, self.private_name, values)

    def __delete__(self, instance):
        if instance is not None:
            delattr(instance, self.private_name)

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name



from __future__ import annotations

from collections.abc import Mapping, Sequence

import numpy as np
import pandas as pd
import sharrow.dataset as sd
import xarray as xr
from sharrow.dataset import construct

from .dim_names import ALTID, ALTIDX, CASEALT, CASEID, CASEPTR


def _steal(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper.__doc__ = func.__doc__
    wrapper.__name__ = func.__name__
    return staticmethod(wrapper)


def _initialize_for_larch(obj, caseid=None, alts=None):
    """
    Initialize a Dataset for use with larch.

    Parameters
    ----------
    obj : Dataset
        The dataset being initialized.
    caseid : str, optional
        The name of a dimension referencing cases.
    alts : Mapping or str or array-like, optional
        If given as a mapping, links alternative codes to names.
        A string names a dimension that defines the alternatives.
        An array or list of integers gives codes for the alternatives,
        which are otherwise unnamed.

    Returns
    -------
    Dataset
    """
    if caseid is not None:
        if caseid not in obj.sizes:
            raise ValueError(f"no dim named '{caseid}' to make into {CASEID}")
        obj.attrs[CASEID] = caseid
    if isinstance(alts, pd.Series):
        alts_dim_name = alts.name or alts.index.name or "_altid_"
        alts_k = xr.DataArray(
            alts.index,
            dims=alts_dim_name,
        )
        alts_v = xr.DataArray(
            alts.values,
            dims=alts_dim_name,
        )
    elif isinstance(alts, Mapping):
        alts_dim_name = "_altid_"
        alts_k = xr.DataArray(
            list(alts.keys()),
            dims=alts_dim_name,
        )
        alts_v = xr.DataArray(
            list(alts.values()),
            dims=alts_dim_name,
        )
    elif isinstance(alts, str):
        alts_dim_name = alts
        alts_k = alts_v = None
    elif alts is None:
        alts_dim_name = None
        alts_k = alts_v = None
    else:
        alts_dim_name = getattr(alts, "name", "_altid_") or "_altid_"
        alts_v = np.asarray(alts).reshape(-1)
        alts_k = None
    if alts_dim_name:
        obj.attrs[ALTID] = alts_dim_name
    if alts_k is not None:
        if np.issubdtype(alts_v, np.integer) and not np.issubdtype(alts_k, np.integer):
            obj.coords[alts_dim_name] = alts_v
            obj.coords["alt_names"] = alts_k
        else:
            obj.coords[alts_dim_name] = alts_k
            obj.coords["alt_names"] = alts_v
    elif alts_v is not None:
        obj.coords[alts_dim_name] = alts_v
    return obj


def initialize_for_larch(
    obj, caseid=None, alts: Mapping[int, str] | str | Sequence[int] | None = None
) -> xr.Dataset:
    """
    Initialize a Dataset for use with larch.

    Parameters
    ----------
    obj : Dataset
        The dataset being initialized.
    caseid : str, optional
        The name of a dimension referencing cases.
    alts : Mapping or str or array-like, optional
        If given as a mapping, links alternative codes to names.
        A string names a dimension that defines the alternatives.
        An array or list of integers gives codes for the alternatives,
        which are otherwise unnamed.

    Returns
    -------
    Dataset
    """
    result = construct(obj)
    result = _initialize_for_larch(result, caseid, alts)
    return result


@xr.register_dataset_accessor("construct")
class _DatasetConstruct:
    _parent_class = xr.Dataset

    def __new__(cls, *args, **kwargs):
        if kwargs:
            source = construct(*args)
            return _initialize_for_larch(source, **kwargs)
        else:
            return super().__new__(cls, *args, **kwargs)

    def __init__(self, x=None, **kwargs):
        self._obj = x

    @classmethod
    def __call__(cls, source, caseid=None, alts=None):
        """
        Construct a Dataset from various similar objects.

        Parameters
        ----------
        source : pandas.DataFrame, pyarrow.Table, xarray.Dataset, or Sequence[str]
            The source from which to create a Dataset.  DataFrames and Tables
            are converted to Datasets that have one dimension (the rows) and
            seperate variables for each of the columns.  A list of strings
            creates a dataset with those named empty variables.
        caseid : str, optional
            The name of a dimension referencing cases.
        alts : Mapping or str or array-like, optional
            If given as a mapping, links alternative codes to names.
            A string names a dimension that defines the alternatives.
            An array or list of integers gives codes for the alternatives,
            which are otherwise unnamed.

        Returns
        -------
        Dataset
        """
        source = construct(source)
        return _initialize_for_larch(source, caseid, alts)

    from_omx = _steal(sd.from_omx)
    from_omx_3d = _steal(sd.from_omx_3d)

    @classmethod
    def from_idco(cls, df, alts=None):
        """
        Construct a Dataset from an idco-format DataFrame.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idco-format DataFrame, with
            the caseid's in a single-level index,
        alts : Mapping or array-like, optional
            If given as a mapping, links alternative codes to names.
            An array or list of integers gives codes for the alternatives,
            which are otherwise unnamed.

        Returns
        -------
        Dataset
        """
        if df.index.nlevels != 1:
            raise ValueError(
                "source idco dataframe must have a one level Index giving case id's"
            )
        caseidname = df.index.name or "index"
        ds = cls()(df, caseid=caseidname, alts=alts)
        ds = ds.dc.set_dtypes(df)
        return ds

    @classmethod
    def from_idca(
        cls, df, crack=True, altnames=None, avail="_avail_", fill_missing=None
    ):
        """
        Construct a Dataset from an idca-format DataFrame.

        This method loads the data as dense arrays.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idca-format or idce-format DataFrame,
            with the caseid's and altid's in a two-level pandas MultiIndex.
        crack : bool, default True
            If True, the `dissolve_zero_variance` method is applied before
            repairing dtypes, to ensure that missing value are handled
            properly.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.
        avail : str, default '_avail_'
            When the imported data is in idce format (i.e. sparse) then
            an availability indicator is computed and given this name.
        fill_missing : scalar or Mapping, optional
            Fill values to use for missing values when imported data is
            in idce format (i.e. sparse).  Give a single value to use
            globally, or a mapping of {variable: value} or {dtype: value}.

        Returns
        -------
        Dataset

        See Also
        --------
        Dataset.from_idce : Construct a Dataset from a sparse idca-format DataFrame.

        """
        if df.index.nlevels != 2:
            raise ValueError(
                "source idca dataframe must have a two "
                "level MultiIndex giving case and alt id's"
            )
        caseidname, altidname = df.index.names

        # check altids are integers, if they are not then fix it
        if df.index.levels[1].dtype.kind != "i":
            if altnames is None:
                altnames = df.index.levels[1]
                df.index = df.index.set_levels(np.arange(1, len(altnames) + 1), level=1)
            else:
                new_index = df.index.get_level_values(1).astype(
                    pd.CategoricalDtype(altnames)
                )
                df.index = df.index.set_codes(new_index.codes, level=1).set_levels(
                    np.arange(1, len(altnames) + 1), level=1
                )

        ds = cls()(df, caseid=caseidname, alts=altidname)
        if crack:
            ds = ds.dc.dissolve_zero_variance()
        ds = ds.dc.set_dtypes(df)
        if altnames is not None:
            ds = ds.dc.set_altnames(altnames)
        if avail not in ds and len(df) < ds.dc.n_cases * ds.dc.n_alts:
            av = (
                xr.DataArray.from_series(pd.Series(1, index=df.index))
                .fillna(0)
                .astype(np.int8)
            )
            ds[avail] = av
            if fill_missing is not None:
                if isinstance(fill_missing, Mapping):
                    for k, i in ds.items():
                        if ds.dc.ALTID not in i.dims:
                            continue
                        if k not in fill_missing and i.dtype not in fill_missing:
                            continue
                        filler = fill_missing.get(k, fill_missing[i.dtype])
                        ds[k] = i.where(ds["_avail_"] != 0, filler)
                else:
                    for k, i in ds.items():
                        if ds.dc.ALTID not in i.dims:
                            continue
                        ds[k] = i.where(ds["_avail_"] != 0, fill_missing)
        return ds

    @classmethod
    def new_idca(self, caseids, altids, altnames=None):
        """
        Create an empty idca Dataset by populating it with case and alt ids.

        Parameters
        ----------
        caseids : array-like
            The case ids to use.
        altids : array-like
            The alternative ids to use.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.

        Returns
        -------
        Dataset
        """
        objs = {}
        try:
            caseid_name = caseids.name or "caseid"
        except AttributeError:
            caseid_name = "caseid"
        objs[caseid_name] = (
            caseids if not isinstance(caseids, pd.Series) else caseids.values
        )

        try:
            altid_name = altids.name or "altid"
        except AttributeError:
            altid_name = "altid"
        objs[altid_name] = (
            altids if not isinstance(altids, pd.Series) else altids.values
        )

        result = xr.Dataset(objs)
        _initialize_for_larch(result, caseid_name, altid_name)
        if altnames is not None:
            result = result.dc.set_altnames(altnames)
        return result

    @classmethod
    def from_idce(
        cls,
        df,
        crack=True,
        altnames=None,
        dim_name=None,
        alt_index="alt_idx",
        case_index=None,
        case_pointer=None,
    ):
        """
        Construct a Dataset from a sparse idca-format DataFrame.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idca-format or idce-format DataFrame,
            with the caseid's and altid's in a two-level pandas MultiIndex.
        crack : bool, default False
            If True, the `dissolve_zero_variance` method is applied before
            repairing dtypes, to ensure that missing value are handled
            properly.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.
        dim_name : str, optional
            Name to apply to the sparse index dimension.
        alt_index : str, default 'alt_idx'
            Add the alt index (position) for each sparse data row as a
            coords array with this name.
        case_index : str, optional
            Add the case index (position) for each sparse data row as a
            coords array with this name. If not given, this array is not
            stored but it can still be reconstructed later from the case
            pointers.
        case_pointer : str, optional
            Use this name for the case_ptr dimension, overriding the
            default.

        Returns
        -------
        Dataset

        See Also
        --------
        Dataset.from_idca : Construct a dense Dataset from a idca-format DataFrame.
        """
        if df.index.nlevels != 2:
            raise ValueError(
                "source idce dataframe must have a two "
                "level MultiIndex giving case and alt id's"
            )
        caseidname, altidname = df.index.names
        caseidname = caseidname or CASEID
        altidname = altidname or ALTID
        dim_name = dim_name or CASEALT
        case_pointer = case_pointer or CASEPTR
        ds = cls()(df.reset_index(drop=True).rename_axis(index=dim_name))
        ds.coords[caseidname] = xr.DataArray(df.index.levels[0], dims=caseidname)
        ds.coords[altidname] = xr.DataArray(df.index.levels[1], dims=altidname)
        if case_index is not None:
            ds.coords[case_index] = xr.DataArray(df.index.codes[0], dims=dim_name)
        if alt_index is None:
            raise ValueError("alt_index cannot be None")
        ds.coords[alt_index] = xr.DataArray(df.index.codes[1], dims=dim_name)
        ds.coords[case_pointer] = xr.DataArray(
            np.where(np.diff(df.index.codes[0], prepend=np.nan, append=np.nan))[0],
            dims=case_pointer,
        )
        ds.attrs["_exclude_dims_"] = (caseidname, altidname, case_pointer)
        ds.attrs[CASEID] = caseidname
        ds.attrs[ALTID] = altidname
        ds.attrs[CASEALT] = dim_name
        ds.attrs[ALTIDX] = alt_index
        ds.attrs[CASEPTR] = case_pointer
        ds = ds.drop_vars(dim_name)
        if crack:
            ds = ds.dc.dissolve_zero_variance()
        ds = ds.dc.set_dtypes(df)
        if altnames is not None:
            ds = ds.dc.set_altnames(altnames)
        return ds



from __future__ import annotations

import numpy
import pandas

from ..data_services.h5.h5pod.generic import CArray


def tally(arr):
    if isinstance(arr, CArray):
        arr_ = arr
        arr = arr[:]
        values, freqs = numpy.unique(arr[~pandas.isnull(arr)], return_counts=True)
        try:
            d = arr_.DICTIONARY
        except Exception:
            d = {}
        s = pandas.Series(freqs, index=[d.get(i, i) for i in values]).sort_values(
            ascending=False
        )
        nan_count = pandas.isnull(arr).sum()
        if nan_count:
            s[numpy.NaN] = nan_count
        return s
    else:
        values, freqs = numpy.unique(arr[~pandas.isnull(arr)], return_counts=True)
        s = pandas.Series(freqs, index=values).sort_values(ascending=False)
        nan_count = pandas.isnull(arr).sum()
        if nan_count:
            s[numpy.NaN] = nan_count
        return s


def tally_weighted(arr, wgt):
    if isinstance(arr, CArray):
        raise NotImplementedError()
    else:
        arr = arr.reshape(-1)
        wgt = wgt.reshape(-1)

        values, inv = numpy.unique(arr[~pandas.isnull(arr)], return_inverse=True)
        s = pandas.Series(numpy.bincount(inv, wgt), index=values).sort_values(
            ascending=False
        )
        nan_count = wgt[pandas.isnull(arr)].sum()
        if nan_count:
            s[numpy.NaN] = nan_count
        return s



from __future__ import annotations

from .._optional import jax


def keysplit(key, shapes):
    if key.ndim == 1:
        new_keys = jax.random.split(key, shapes[0])
    else:
        new_keys = jax.vmap(jax.random.split, in_axes=(0, None))(key, shapes[0])
    return new_keys, shapes[1:]



from __future__ import annotations

import contextlib
import os
import sys


@contextlib.contextmanager
def shush(stdout=False, stderr=True):
    """Suppress stdout and/or stderr using a context manager."""
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        if stdout:
            sys.stdout = devnull
        if stderr:
            sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr



from __future__ import annotations

import numpy


class NonUniqueLookup(Exception):
    pass


def is_sorted_and_unique(arr):
    arr = arr.reshape(-1)
    return numpy.all(arr[:-1] < arr[1:])


def is_sorted(arr):
    arr = arr.reshape(-1)
    return numpy.all(arr[:-1] <= arr[1:])


def label_to_index(labels, arr):
    """Convert an array of lookup-able values into indexes.

    If you have an array of lookup-able values (e.g., TAZ identifiers) and you
    want to convert them to 0-based indexes for use in accessing matrix data,
    this is the function for you.

    Parameters
    ----------
    labels : 1d array-like
            An array of labels.
    arr : array-like
            An array of values that appear in the label array. This method uses
            numpy.digitize to process values, so any target value that appears in `arr` but does
            not appear in the labels will be assigned the index of the smallest label
            value that is greater than the target, or the maximum label value if no label value
            is greater than the target.

    Returns
    -------
    array
            An array of index (int) values, with the same shape as `arr`.

    Raises
    ------
    OMXNonUniqueLookup
            When the lookup does not contain a set of unique values, this tool is not appropriate.

    """
    try:
        if len(labels) == 0:
            return labels
    except TypeError:
        return labels
    labels = numpy.asarray(labels)
    if is_sorted_and_unique(labels):
        return numpy.digitize(arr, labels, right=True)
    uniq_labels, uniq_indexes = numpy.unique(labels, return_inverse=True)
    if len(uniq_labels) != len(labels):
        raise NonUniqueLookup("lookup does not have unique labels for each item")
    index_malordered = numpy.digitize(arr, uniq_labels, right=True)
    return uniq_indexes[index_malordered]


def labels_to_unique_ids(various_labels, lowest_label_number=1):
    """Convert an array or list of various labels to unique sequential integer code numbers.

    Parameters
    ----------
    various_labels : array or list
            A bunch of various labels to re-map.  Duplicates are allowed.
    lowest_label_number : int
            The lowest label number returned, by default 1.

    Returns
    -------
    array
            An array of unique labels, in order.
    ids
            An array of integer code numbers, with the same shape as the
            input `various_labels`.

    """
    ordered_labels, label_numbers = numpy.unique(various_labels, return_inverse=True)
    label_numbers += lowest_label_number
    return ordered_labels, label_numbers


def orthogonal_unit_vector(nDims, positionOne, dtype=numpy.float64):
    z = numpy.zeros(nDims, dtype=dtype)
    z[positionOne] = 1
    return z


def is_all_integer(arr):
    if arr.size < 100:
        if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
            return True
    else:
        if numpy.all(numpy.equal(numpy.mod(arr.ravel()[:100], 1), 0)):
            if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
                return True
    return False


def is_all_integer_or_nan(arr):
    arr_nan = arr[~numpy.isnan(arr)]
    if arr_nan.size < 100:
        if numpy.all(numpy.equal(numpy.mod(arr_nan, 1), 0)):
            return True
    else:
        if numpy.all(numpy.equal(numpy.mod(arr_nan.ravel()[:100], 1), 0)):
            if numpy.all(numpy.equal(numpy.mod(arr_nan, 1), 0)):
                return True
    return False


def convert_float_to_int_if_lossless(arr, inttype=numpy.int32):
    if arr.size < 100:
        if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
            return arr.astype(inttype)
    else:
        if numpy.all(numpy.equal(numpy.mod(arr.ravel()[:100], 1), 0)):
            if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
                return arr.astype(inttype)
    return arr


def failable_iter_to_set(iterable, transformer):
    s = set()
    for i in iterable:
        try:
            s.add(transformer(i))
        except AttributeError:
            pass
    return s


def failable_iter_to_unique(iterable, transformer):
    s_cache = None
    for i in iterable:
        try:
            s = transformer(i)
        except AttributeError:
            pass
        else:
            if s_cache is not None:
                if s_cache != s:
                    return None
            else:
                s_cache = s
    return s_cache


def unique_successful_transform(iterable, transformer, accept_longest=False):
    s = failable_iter_to_unique(iterable, transformer)
    if len(s) == 1:
        return s.pop()
    if accept_longest:
        candidate = None
        candidate_len = 0
        for i in s:
            if len(str(i)) > candidate_len:
                candidate_len = len(str(i))
                candidate = i
        return candidate
    return None


def scalarize(a):
    try:
        return a.item()
    except (ValueError, AttributeError):
        return a



from __future__ import annotations

import contextlib
import warnings


@contextlib.contextmanager
def ignore_warnings():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield


class LarchWarning(Warning):
    """Base class for all Larch warnings."""


class NoDefaultValueWarning(LarchWarning):
    pass


class WeightRescaleWarning(Warning):
    """Warning issued when weights are rescaled."""


def no_default_value(message):
    # warnings.warn(message, NoDefaultValueWarning, stacklevel=2)
    pass



def example(extract='m', estimate=False):
    import larch

    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    d = d.sel(caseid=d.femdum == 1)

    m = larch.Model(d, compute_engine="numba")

    from larch import P, X

    for a in [4, 5, 6]:
        m.utility_co[a] += X("hhinc") * P(f"hhinc#{a}")

    for i in d["alt_names"][1:3]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_SR")
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    for i in d["alt_names"][3:]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_" + name)
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    m.utility_ca = (
        +X("totcost/hhinc") * P("costbyincome")
        + X("tottime * (altid <= 4)") * P("motorized_time")
        + X("tottime * (altid >= 5)") * P("nonmotorized_time")
        + X("ovtt/dist * (altid <= 4)") * P("motorized_ovtbydist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 17D, Segmented for females"

    m.choice_avail_summary()

    m.set_cap(25)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, options={"maxiter": 1000, "ftol": 1e-10})
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Zonal",
            "wkcbd.*",
            "wkempden.*",
        ),
        (
            "Household",
            "hhinc.*",
            "vehbywrk.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_Bike": -1.0412780932548384,
        "ASC_SR2": -1.5539810652068256,
        "ASC_SR3+": -3.1854305680124058,
        "ASC_Transit": -0.46976829166153894,
        "ASC_Walk": 1.3688704069415694,
        "costbyincome": -0.043530473293242354,
        "hhinc#4": -0.008929802961855733,
        "hhinc#5": -0.03840955269802502,
        "hhinc#6": -0.005082499964774252,
        "motorized_ovtbydist": -0.08879718827392451,
        "motorized_time": -0.01913530894108756,
        "nonmotorized_time": -0.0712577146848601,
        "vehbywrk_Bike": -0.10343275503303231,
        "vehbywrk_SR": -0.6135515751452322,
        "vehbywrk_Transit": -1.1794317963027092,
        "vehbywrk_Walk": -0.917891133747879,
        "wkcbd_Bike": 1.0534091265461347,
        "wkcbd_SR2": 0.45194163205040755,
        "wkcbd_SR3+": 0.3685888989631644,
        "wkcbd_Transit": 1.3811852123386674,
        "wkcbd_Walk": -0.008517015595822975,
        "wkempden_Bike": 0.004039748186460425,
        "wkempden_SR2": 0.0029908432533657533,
        "wkempden_SR3+": 0.005134839725445598,
        "wkempden_Transit": 0.004565945257271104,
        "wkempden_Walk": 0.005465423507422594,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(expected_x[k], 2e-2), (
            f"{k}, {revealed_x[k] / expected_x[k]}"
        )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import ast

import numpy


def inXd(a, *b, **c):
    return numpy.in1d(a, *b, **c).reshape(a.shape)


class AstWrapper(ast.NodeTransformer):
    def visit_Compare(self, node):
        """Change the `in` operator to numpy.in1d."""
        if len(node.ops) == 1 and isinstance(node.ops[0], ast.In):
            in1d = ast.Name(id="inXd", ctx=ast.Load())
            return ast.Call(in1d, [node.left, node.comparators[0]], [])
        elif len(node.ops) == 1 and isinstance(node.ops[0], ast.NotIn):
            in1d = ast.Name(id="inXd", ctx=ast.Load())
            return ast.Call(
                in1d,
                [node.left, node.comparators[0]],
                [ast.keyword("invert", ast.NameConstant(True))],
            )
        else:
            return node


def asterize(cmd, mode="eval"):
    tree = ast.parse(cmd, mode=mode)
    tree = AstWrapper().visit(tree)
    # Add lineno & col_offset to the nodes we created
    ast.fix_missing_locations(tree)
    co = compile(tree, "<ast>", mode)
    return co



from __future__ import annotations

import os
import sys

from .. import __version__


class Info:
    def __init__(
        self, appname="Larch", extra=True, version=None, path=None, minimal=False
    ):
        self.appname = appname
        self.extra = extra
        self.version = version or __version__
        self.minimal = minimal
        from .. import __path__

        self.path = path or __path__[0]

    def __repr__(self):
        r = f"┌── {self.appname.upper()} {self.version} " + "─" * (
            57 - len(self.version)
        )
        v = "\n│".join(sys.version.split("\n"))
        r += f"\n│Python {v}"
        r += f"\n│EXE ─ {sys.executable}"
        r += f"\n│CWD ─ {os.getcwd()}"
        for p in sys.path[:1]:
            r += f"\n│PTH ┬ {p}"
        for p in sys.path[1:-1]:
            r += f"\n│    ├ {p}"
        for p in sys.path[-1:]:
            r += f"\n│    └ {p}"
        r += "\n└───────────────────────────────────────────────────────────────────────────"
        return r

    def _repr_html_(self):
        from ..util.styles import _default_css_jupyter

        style_prefix = f"<style>{_default_css_jupyter}</style>\n"
        from xmle import Elem

        xsign = Elem("div", {"class": "larch_head_tag"})
        from .images import favicon

        p = xsign.elem("p", {"style": "margin-top:6px"})
        p.elem(
            "img",
            {
                "width": "32",
                "height": "32",
                "src": f"data:image/png;base64,{favicon}",
                "style": "float:left;position:relative;top:-3px;padding-right:0.2em;",
            },
            tail=f" {self.appname} ",
        )
        p.elem("span", {"class": "larch_head_tag_ver"}, text=f" {self.version} ")
        if not self.minimal:
            p.elem("span", {"class": "larch_head_tag_pth"}, text=f" {self.path} ")

            if self.extra:
                v = "\n│".join(sys.version.split("\n"))
                # xsign.elem("br")
                xinfo = xsign.elem(
                    "div",
                    {
                        "class": "larch_head_tag_more",
                        "style": "padding:7px",
                    },
                    text=f"Python {v}",
                )
                xinfo.elem("br", tail=f"EXE - {sys.executable}")
                xinfo.elem("br", tail=f"CWD - {os.getcwd()}")
                xinfo.elem("br", tail="PATH - ")
                ul = xinfo.elem("ul", {"style": "margin-top:0; margin-bottom:0;"})
                for p in sys.path:
                    ul.elem("li", text=p)

        return style_prefix + xsign.tostring()

    def __call__(self, *args, **kwargs):
        """
        Do nothing (for now).

        Returns
        -------
        self
        """
        return self


def ipython_status(magic_matplotlib=False):
    message_set = set()
    try:
        # This will work in iPython, and fail otherwise
        cfg = get_ipython().config  # noqa: F821
    except Exception:
        message_set.add("Not IPython")
    else:
        import IPython
        import IPython.core.error

        message_set.add("IPython")

        if magic_matplotlib:
            try:
                get_ipython().magic("matplotlib inline")  # noqa: F821
            except (IPython.core.error.UsageError, KeyError):
                message_set.add("IPython inline plotting not available")

        # Caution: cfg is an IPython.config.loader.Config
        if cfg["IPKernelApp"]:
            message_set.add("IPython QtConsole")
            try:
                if cfg["IPKernelApp"]["pylab"] == "inline":
                    message_set.add("pylab inline")
                else:
                    message_set.add("pylab loaded but not inline")
            except Exception:
                message_set.add("pylab not loaded")
        elif cfg["TerminalIPythonApp"]:
            try:
                if cfg["TerminalIPythonApp"]["pylab"] == "inline":
                    message_set.add("pylab inline")
                else:
                    message_set.add("pylab loaded but not inline")
            except Exception:
                message_set.add("pylab not loaded")
    return message_set



from __future__ import annotations

_color_rgb256 = {}
_color_rgb256["sky"] = (35, 192, 241)
_color_rgb256["ocean"] = (29, 139, 204)
_color_rgb256["night"] = (100, 120, 186)
_color_rgb256["forest"] = (39, 182, 123)
_color_rgb256["lime"] = (128, 189, 1)
_color_rgb256["orange"] = (246, 147, 0)
_color_rgb256["red"] = (246, 1, 0)


def hexcolor(color):
    c = _color_rgb256[color.casefold()]
    return "#{}{}{}".format(
        *(hex(c[i])[-2:] if c[i] > 15 else "0" + hex(c[i])[-1:] for i in range(3))
    )


def strcolor_rgb256(color):
    c = _color_rgb256[color.casefold()]
    return "{},{},{}".format(*c)



from __future__ import annotations

from datetime import datetime


def timesize_single(t):
    if t < 60:
        return f"{t:.2f}s"
    elif t < 3600:
        return f"{t / 60:.2f}m"
    elif t < 86400:
        return f"{t / 3600:.2f}h"
    else:
        return f"{t / 86400:.2f}d"


def timesize_stack(t):
    if t < 60:
        return f"{t:.2f}s"
    elif t < 3600:
        return f"{t // 60:.0f}m {timesize_stack(t % 60)}"
    elif t < 86400:
        return f"{t // 3600:.0f}h {timesize_stack(t % 3600)}"
    else:
        return f"{t // 86400:.0f}d {timesize_stack(t % 86400)}"


class Seconds(float):
    def __repr__(self):
        t = self
        if t < 60:
            return f"{t:.2f}s"
        elif t < 3600:
            return f"{t // 60:.0f}m {timesize_stack(t % 60)}"
        elif t < 86400:
            return f"{t // 3600:.0f}h {timesize_stack(t % 3600)}"
        else:
            return f"{t // 86400:.0f}d {timesize_stack(t % 86400)}"


class Timer:
    def __init__(self):
        self.restart()

    def restart(self):
        self._starttime = datetime.now()
        self._stoptime = None
        return self._starttime

    def start(self):
        return self._starttime

    def stop(self):
        self._stoptime = datetime.now()
        return self._stoptime

    def elapsed(self):
        if self._stoptime:
            return self._stoptime - self._starttime
        else:
            return datetime.now() - self._starttime

    def __enter__(self):
        self.restart()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



from __future__ import annotations


class Nothing:
    def __getattr__(self, name):
        return Nothing()

    def __call__(self, *args, **kwargs):
        return Nothing()

    def __bool__(self):
        return False


# optional import: JAX
try:
    import jax
    import jax.numpy as jnp
    import jax.scipy as js
except ImportError:
    jax = Nothing()
    jnp = Nothing()
    js = Nothing()
    print("JAX not found. Some functionality will be unavailable.")

# optional import: matplotlib
try:
    from matplotlib import pyplot
except ImportError:
    pyplot = Nothing()


__all__ = [
    "jax",
    "jnp",
    "js",
    "pyplot",
]



from __future__ import annotations

default_config = """---

rules:
  braces:
    min-spaces-inside: 0
    max-spaces-inside: 0
    min-spaces-inside-empty: -1
    max-spaces-inside-empty: -1
  brackets:
    min-spaces-inside: 0
    max-spaces-inside: 0
    min-spaces-inside-empty: -1
    max-spaces-inside-empty: -1
  colons:
    max-spaces-before: 0
    max-spaces-after: 999
  commas:
    max-spaces-before: 0
    min-spaces-after: 1
    max-spaces-after: 1
  comments:
    level: warning
    require-starting-space: false
    min-spaces-from-content: 1
  comments-indentation:
    level: warning
  document-end: disable
  document-start: disable
  #  level: warning
  #  present: true
  empty-lines:
    max: 99
    max-start: 99
    max-end: 99
  empty-values:
    forbid-in-block-mappings: false
    forbid-in-flow-mappings: false
  hyphens:
    max-spaces-after: 1
  indentation:
    spaces: consistent
    indent-sequences: true
    check-multi-line-strings: false
  key-duplicates: enable
  key-ordering: disable
  line-length:
    max: 880
    allow-non-breakable-words: true
    allow-non-breakable-inline-mappings: false
  new-line-at-end-of-file: enable
  new-lines:
    type: unix
  octal-values:
    forbid-implicit-octal: false
    forbid-explicit-octal: false
  trailing-spaces: disable
  truthy: disable
  #  level: warning
"""


class Format:
    @staticmethod
    def parsable(problem, filename):
        return f"{filename}:{problem.line}:{problem.column}: [{problem.level}] {problem.message}"

    @staticmethod
    def standard(problem, filename):
        line = f"  {problem.line:d}:{problem.column:d}"
        line += max(12 - len(line), 0) * " "
        line += problem.level
        line += max(21 - len(line), 0) * " "
        line += problem.desc
        if problem.rule:
            line += f"  ({problem.rule:s})"
        return line

    @staticmethod
    def standard_color(problem, filename):
        line = f"  \033[2m{problem.line:d}:{problem.column:d}\033[0m"
        line += max(20 - len(line), 0) * " "
        if problem.level == "warning":
            line += f"\033[33m{problem.level}\033[0m"
        else:
            line += f"\033[31m{problem.level}\033[0m"
        line += max(38 - len(line), 0) * " "
        line += problem.desc
        if problem.rule:
            line += f"  \033[2m({problem.rule})\033[0m"
        return line


def yaml_check(file, config_file=None, logger=None, encoding="utf-8"):
    if logger is None:
        log = print
    else:
        log = logger.error

    try:
        from yamllint import linter
        from yamllint.config import YamlLintConfig
        from yamllint.linter import PROBLEM_LEVELS
    except ImportError:
        log(
            "yamllint is not installed, cannot inspect yaml file for formatting quality"
        )
    else:
        filepath = file[2:] if file.startswith("./") else file

        if config_file is not None:
            conf = YamlLintConfig(file=config_file)
        else:
            conf = YamlLintConfig(content=default_config)

        first = True
        any_errors = False
        max_level = 0
        with open(filepath, encoding=encoding) as f:
            for problem in linter.run(f, conf, filepath):
                if first:
                    log(f"FOUND YAML ERRORS IN {file}")
                    first = False
                    any_errors = True

                log(Format.standard(problem, file))

                max_level = max(max_level, PROBLEM_LEVELS[problem.level])

            if any_errors:
                log(f"END OF YAML ERRORS IN {file}")



from __future__ import annotations

import re
import warnings

import numpy as np
import pandas as pd
import sharrow as sh
import xarray as xr
from pandas.errors import UndefinedVariableError
from xarray.core import dtypes

from . import construct as construct
from . import flow as flow
from . import patch as patch
from .dim_names import ALTID as _ALTID
from .dim_names import ALTIDX as _ALTIDX
from .dim_names import CASEALT as _CASEALT
from .dim_names import CASEID as _CASEID
from .dim_names import CASEPTR as _CASEPTR
from .patch import register_dataarray_classmethod

# from .dim_names import GROUPID as _GROUPID
# from .dim_names import INGROUP as _INGROUP

try:
    from sharrow import DataArray as _sharrow_DataArray
    from sharrow import Dataset as _sharrow_Dataset
    from sharrow import DataTree as _sharrow_DataTree
    from sharrow.accessors import register_dataarray_method
except ImportError:
    warnings.warn("larch.dataset requires the sharrow library", stacklevel=2)

    class _noclass:
        pass

    _sharrow_Dataset = xr.Dataset
    _sharrow_DataArray = xr.DataArray
    _sharrow_DataTree = _noclass
    register_dataarray_method = lambda x: x


DataArray = _sharrow_DataArray


@register_dataarray_classmethod
def zeros(cls, *coords, dtype=np.float64, name=None, attrs=None):
    """
    Construct a dataset filled with zeros.

    Parameters
    ----------
    coords : Tuple[array-like]
        A sequence of coordinate vectors.  Ideally each should have a
        `name` attribute that names a dimension, otherwise placeholder
        names are used.
    dtype : dtype, default np.float64
        dtype of the new array. If omitted, it defaults to np.float64.
    name : str or None, optional
        Name of this array.
    attrs : dict_like or None, optional
        Attributes to assign to the new instance. By default, an empty
        attribute dictionary is initialized.

    Returns
    -------
    DataArray
    """
    dims = []
    shape = []
    coo = {}
    for n, c in enumerate(coords):
        i = getattr(c, "name", f"dim_{n}")
        dims.append(i)
        shape.append(len(c))
        coo[i] = c
    return cls(
        data=np.zeros(shape, dtype=dtype),
        coords=coo,
        dims=dims,
        name=name,
        attrs=attrs,
    )


@register_dataarray_classmethod
def ones(cls, *coords, dtype=np.float64, name=None, attrs=None):
    """
    Construct a dataset filled with ones.

    Parameters
    ----------
    coords : Tuple[array-like]
        A sequence of coordinate vectors.  Ideally each should have a
        `name` attribute that names a dimension, otherwise placeholder
        names are used.
    dtype : dtype, default np.float64
        dtype of the new array. If omitted, it defaults to np.float64.
    name : str or None, optional
        Name of this array.
    attrs : dict_like or None, optional
        Attributes to assign to the new instance. By default, an empty
        attribute dictionary is initialized.

    Returns
    -------
    DataArray
    """
    dims = []
    shape = []
    coo = {}
    for n, c in enumerate(coords):
        i = getattr(c, "name", f"dim_{n}")
        dims.append(i)
        shape.append(len(c))
        coo[i] = c
    return cls(
        data=np.ones(shape, dtype=dtype),
        coords=coo,
        dims=dims,
        name=name,
        attrs=attrs,
    )


@register_dataarray_classmethod
def from_zarr(cls, *args, name=None, **kwargs):
    dataset = xr.open_zarr(*args, **kwargs)
    if name is None:
        names = set(dataset.variables) - set(dataset.coords)
        if len(names) == 1:
            name = names.pop()
        else:
            raise ValueError("cannot infer name to load")
    return dataset[name]


@register_dataarray_method
def value_counts(self, index_name="index"):
    """
    Count the number of times each unique value appears in the array.

    Parameters
    ----------
    index_name : str, default 'index'
        Name of index dimension in result.

    Returns
    -------
    DataArray
    """
    values, freqs = np.unique(self, return_counts=True)
    return self.__class__(freqs, dims=index_name, coords={index_name: values})


Dataset = _sharrow_Dataset


class DataTree(_sharrow_DataTree):
    DatasetType = Dataset

    def __init__(
        self,
        graph=None,
        root_node_name=None,
        extra_funcs=(),
        extra_vars=None,
        cache_dir=None,
        relationships=(),
        force_digitization=False,
        **kwargs,
    ):
        super().__init__(
            graph=graph,
            root_node_name=root_node_name,
            extra_funcs=extra_funcs,
            extra_vars=extra_vars,
            cache_dir=cache_dir,
            relationships=relationships,
            force_digitization=force_digitization,
            **kwargs,
        )
        dim_order = []
        c = self.root_dataset.dc.CASEID
        if c is None and len(self.root_dataset.sizes) == 1:
            self.root_dataset.dc.CASEID = list(self.root_dataset.sizes.keys())[0]
        c = self.root_dataset.dc.CASEID
        if c is not None:
            dim_order.append(c)
        a = self.root_dataset.dc.ALTID
        if a is not None:
            dim_order.append(a)
        self.dim_order = tuple(dim_order)

    def idco_subtree(self):
        if "idcoVars" in self.subspaces:
            return self.subspaces["idcoVars"].dc.as_tree()
        return self.drop_dims(self.ALTID, ignore_missing_dims=True)

    @property
    def dc(self):
        return self

    @property
    def CASEID(self):
        """Str : The _caseid_ dimension of the root Dataset."""
        result = self.root_dataset.dc.CASEID
        if result is None:
            warnings.warn("no defined CASEID", stacklevel=2)
            return _CASEID
        return result

    @property
    def ALTID(self):
        """Str : The _altid_ dimension of the root Dataset."""
        result = self.root_dataset.dc.ALTID
        if result is None:
            warnings.warn("no defined ALTID", stacklevel=2)
            return _ALTID
        return result

    @property
    def CASEALT(self):
        """Str : The _casealt_ dimension of the root Dataset, if defined."""
        result = self.root_dataset.attrs.get(_CASEALT, None)
        return result

    @property
    def ALTIDX(self):
        """Str : The _alt_idx_ dimension of the root Dataset, if defined."""
        result = self.root_dataset.attrs.get(_ALTIDX, None)
        return result

    @property
    def CASEPTR(self):
        """Str : The _caseptr_ dimension of the root Dataset, if defined."""
        result = self.root_dataset.attrs.get(_CASEPTR, None)
        return result

    @property
    def n_cases(self):
        """Int : The size of the _caseid_ dimension of the root Dataset."""
        return self.root_dataset.sizes[self.CASEID]

    @property
    def n_alts(self):
        """Int : The size of the _altid_ dimension of the root Dataset."""
        return self.root_dataset.sizes[self.ALTID]

    def query_cases(self, query, parser="pandas", engine=None):
        """
        Return a new DataTree, with a query filter applied to the root Dataset.

        Parameters
        ----------
        query : str
            Python expressions to be evaluated against the data variables
            in the root dataset. The expressions will be evaluated using the pandas
            eval() function, and can contain any valid Python expressions but cannot
            contain any Python statements.
        parser : {"pandas", "python"}, default: "pandas"
            The parser to use to construct the syntax tree from the expression.
            The default of 'pandas' parses code slightly different than standard
            Python. Alternatively, you can parse an expression using the 'python'
            parser to retain strict Python semantics.
        engine : {"python", "numexpr", None}, default: None
            The engine used to evaluate the expression. Supported engines are:

            - None: tries to use numexpr, falls back to python
            - "numexpr": evaluates expressions using numexpr
            - "python": performs operations as if you had eval’d in top level python

        Returns
        -------
        DataTree
            A new DataTree with the same contents as this DataTree, except each
            array of the root Dataset is indexed by the results of the query on
            the CASEID dimension.

        See Also
        --------
        Dataset.query_cases
        """
        obj = self.copy()
        try:
            obj.root_dataset = obj.root_dataset.dc.query_cases(
                query, parser=parser, engine=engine
            )
        except UndefinedVariableError:
            filter = self.idco_subtree().get_expr(
                query, allow_native=False, engine="sharrow", dtype="bool_"
            )
            obj.root_dataset = obj.root_dataset.dc.isel({self.CASEID: filter})
        return obj

    def slice_cases(self, *case_slice):
        if len(case_slice) != 1 or not isinstance(case_slice[0], slice):
            case_slice = slice(*case_slice)
        return self.replace_datasets(
            {self.root_node_name: self.root_dataset.isel({self.CASEID: case_slice})}
        )

    def caseids(self) -> pd.Index:
        """
        Access the caseids coordinates as an index.

        Returns
        -------
        pandas.Index
        """
        try:
            return self.root_dataset.indexes[self.CASEID]
        except KeyError:
            for _k, v in self.subspaces.items():
                if self.CASEID in v.indexes:
                    return v.indexes[self.CASEID]
            raise

    def altids(self) -> pd.Index:
        """
        Access the altids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        try:
            return self.root_dataset.indexes[self.ALTID]
        except KeyError:
            for _k, v in self.subspaces.items():
                if self.ALTID in v.indexes:
                    return v.indexes[self.ALTID]
            raise

    def set_altnames(self, alt_names):
        """
        Set the alternative names for this DataTree.

        Parameters
        ----------
        altnames : Mapping or array-like
            A mapping of (integer) codes to names, or an array or names
            of the same length and order as the alternatives already
            defined in this Dataset.
        """
        self.root_dataset = self.root_dataset.dc.set_altnames(alt_names)

    def alts_mapping(self):
        return self.root_dataset.dc.alts_mapping

    def alts_name_to_id(self):
        return dict((j, i) for (i, j) in self.alts_mapping().items())

    def setup_flow(self, *args, **kwargs) -> sh.Flow:
        """
        Set up a new Flow for analysis using the structure of this DataTree.

        Parameters
        ----------
        definition_spec : dict[str,str]
            Gives the names and expressions that define the variables to
            create in this new `Flow`.
        cache_dir : Path-like, optional
            A location to write out generated python and numba code. If not
            provided, a unique temporary directory is created.
        name : str, optional
            The name of this Flow used for writing out cached files. If not
            provided, a unique name is generated. If `cache_dir` is given,
            be sure to avoid name conflicts with other flow's in the same
            directory.
        dtype : str, default "float32"
            The name of the numpy dtype that will be used for the output.
        boundscheck : bool, default False
            If True, boundscheck enables bounds checking for array indices, and
            out of bounds accesses will raise IndexError. The default is to not
            do bounds checking, which is faster but can produce garbage results
            or segfaults if there are problems, so try turning this on for
            debugging if you are getting unexplained errors or crashes.
        error_model : {'numpy', 'python'}, default 'numpy'
            The error_model option controls the divide-by-zero behavior. Setting
            it to ‘python’ causes divide-by-zero to raise exception like
            CPython. Setting it to ‘numpy’ causes divide-by-zero to set the
            result to +/-inf or nan.
        nopython : bool, default True
            Compile using numba's `nopython` mode.  Provided for debugging only,
            as there's little point in turning this off for production code, as
            all the speed benefits of sharrow will be lost.
        fastmath : bool, default True
            If true, fastmath enables the use of "fast" floating point transforms,
            which can improve performance but can result in tiny distortions in
            results.  See numba docs for details.
        parallel : bool, default True
            Enable or disable parallel computation for certain functions.
        readme : str, optional
            A string to inject as a comment at the top of the flow Python file.
        flow_library : Mapping[str,Flow], optional
            An in-memory cache of precompiled Flow objects.  Using this can result
            in performance improvements when repeatedly using the same definitions.
        extra_hash_data : Tuple[Hashable], optional
            Additional data used for generating the flow hash.  Useful to prevent
            conflicts when using a flow_library with multiple similar flows.
        write_hash_audit : bool, default True
            Writes a hash audit log into a comment in the flow Python file, for
            debugging purposes.
        hashing_level : int, default 1
            Level of detail to write into flow hashes.  Increase detail to avoid
            hash conflicts for similar flows.  Level 2 adds information about
            names used in expressions and digital encodings to the flow hash,
            which prevents conflicts but requires more pre-computation to generate
            the hash.
        dim_exclude : Collection[str], optional
            Exclude these root dataset dimensions from this flow.

        Returns
        -------
        Flow
        """
        if "dim_exclude" not in kwargs:
            if "_exclude_dims_" in self.root_dataset.attrs:
                kwargs["dim_exclude"] = self.root_dataset.attrs["_exclude_dims_"]
        try:
            return super().setup_flow(*args, **kwargs)
        except ValueError as err:
            regex = re.match("^unable to rewrite (.*) to itself$", str(err))
            if regex:
                raise ValueError(
                    f"Setup failed for variable {regex.group(1)}.  Check the expression "
                    f"and the names of the variables in the dataset."
                ) from err
            else:
                raise err


def merge(
    objects,
    compat="no_conflicts",
    join="outer",
    fill_value=dtypes.NA,
    combine_attrs="override",
    *,
    caseid=None,
    alts=None,
):
    """
    Merge any number of xarray objects into a single larch.Dataset as variables.

    Parameters
    ----------
    objects : iterable of Dataset or iterable of DataArray or iterable of dict-like
        Merge together all variables from these objects. If any of them are
        DataArray objects, they must have a name.

    compat : {"identical", "equals", "broadcast_equals", "no_conflicts", "override"}, optional
        String indicating how to compare variables of the same name for
        potential conflicts:
        - "broadcast_equals": all values must be equal when variables are
          broadcast against each other to ensure common dimensions.
        - "equals": all values and dimensions must be the same.
        - "identical": all values, dimensions and attributes must be the
          same.
        - "no_conflicts": only values which are not null in both datasets
          must be equal. The returned dataset then contains the combination
          of all non-null values.
        - "override": skip comparing and pick variable from first dataset

    join : {"outer", "inner", "left", "right", "exact"}, optional
        String indicating how to combine differing indexes in objects.
        - "outer": use the union of object indexes
        - "inner": use the intersection of object indexes
        - "left": use indexes from the first object with each dimension
        - "right": use indexes from the last object with each dimension
        - "exact": instead of aligning, raise `ValueError` when indexes to be
          aligned are not equal
        - "override": if indexes are of same size, rewrite indexes to be
          those of the first object with that dimension. Indexes for the same
          dimension must have the same size in all objects.

    fill_value : scalar or dict-like, optional
        Value to use for newly missing values. If a dict-like, maps
        variable names to fill values. Use a data array's name to
        refer to its values.

    combine_attrs : {"drop", "identical", "no_conflicts", "drop_conflicts", \
                    "override"} or callable, default: "override"
        A callable or a string indicating how to combine attrs of the objects being
        merged:
        - "drop": empty attrs on returned Dataset.
        - "identical": all attrs must be the same on every object.
        - "no_conflicts": attrs from all objects are combined, any that have
          the same name must also have the same value.
        - "drop_conflicts": attrs from all objects are combined, any that have
          the same name but different values are dropped.
        - "override": skip comparing and copy attrs from the first dataset to
          the result.
        If a callable, it must expect a sequence of ``attrs`` dicts and a context object
        as its only parameters.

    caseid : str, optional, keyword only
        This named dimension will be marked as the '_caseid_' dimension.

    alts : str or Mapping or array-like, keyword only
        If given as a str, this named dimension will be marked as the
        '_altid_' dimension.  Otherwise, give a Mapping that defines
        alternative names and (integer) codes or an array of codes.

    Returns
    -------
    Dataset
        Dataset with combined variables from each object.
    """
    return Dataset.construct(
        xr.merge(
            objects,
            compat="no_conflicts",
            join="outer",
            fill_value=dtypes.NA,
            combine_attrs="override",
        ),
        caseid=caseid,
        alts=alts,
    )


# @nb.njit
# def ce_dissolve_zero_variance(ce_data, ce_caseptr):
#     """
#
#     Parameters
#     ----------
#     ce_data : array-like, shape [n_casealts] one-dim only
#     ce_altidx
#     ce_caseptr
#     n_alts
#
#     Returns
#     -------
#     out : ndarray
#     flag : int
#         1 if variance was detected, 0 if no variance was found and
#         the `out` array is valid.
#     """
#     failed = 0
#     if ce_caseptr.ndim == 2:
#         ce_caseptr1 = ce_caseptr[:,-1]
#     else:
#         ce_caseptr1 = ce_caseptr[1:]
#     shape = (ce_caseptr1.shape[0], )
#     out = np.zeros(shape, dtype=ce_data.dtype)
#     c = 0
#     out[0] = ce_data[0]
#     for row in range(ce_data.shape[0]):
#         if row == ce_caseptr1[c]:
#             c += 1
#             out[c] = ce_data[row]
#         else:
#             if out[c] != ce_data[row]:
#                 failed = 1
#                 break
#     return out, failed


# @nb.njit
# def case_ptr_to_indexes(n_casealts, case_ptrs):
#     case_index = np.zeros(n_casealts, dtype=np.int64)
#     for c in range(case_ptrs.shape[0]-1):
#         case_index[case_ptrs[c]:case_ptrs[c + 1]] = c
#     return case_index



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import PX, P, X

    m.utility_co[2] = P("ASC_SR2") + P("hhinc#2,3") * X("hhinc")
    m.utility_co[3] = P("ASC_SR3P") + P("hhinc#2,3") * X("hhinc")
    m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
    m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
    m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")

    m.utility_ca = (
        +P("nonmotorized_time") * X("(altid>4) * tottime")
        + P("motorized_ovtt") * X("(altid <= 4) * ovtt")
        + P("motorized_ivtt") * X("(altid <= 4) * ivtt")
        + PX("totcost")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 6, Motorized Times"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, method="bhhh")
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            "totcost",
            "nonmotorized_time",
            "motorized_ivtt",
            "motorized_ovtt",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_BIKE": -1.719093452086421,
        "ASC_SR2": -2.4299943985819956,
        "ASC_SR3P": -3.8834310483961114,
        "ASC_TRAN": -0.4898195783250097,
        "ASC_WALK": 0.4091981983094832,
        "hhinc#2,3": -0.0015835444926238225,
        "hhinc#4": -0.005692708506297093,
        "hhinc#5": -0.01221964685106249,
        "hhinc#6": -0.009303042887650576,
        "motorized_ivtt": -0.002540641825564394,
        "motorized_ovtt": -0.07594330410291915,
        "nonmotorized_time": -0.06319642667912634,
        "totcost": -0.0047982963933021535,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(expected_x[k], 2e-2), (
            f"{k}, {revealed_x[k] / expected_x[k]}"
        )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import numpy as np

    import larch as lx
    from larch import P, X

    hh, pp, tour, skims = lx.example(200, ["hh", "pp", "tour", "skims"])

    from addicty import Dict

    Mode = Dict(
        DA=1,
        SR=2,
        Walk=3,
        Bike=4,
        Transit=5,
    ).freeze()

    tour_dataset = lx.Dataset.construct.from_idco(tour.set_index("TOURID"), alts=Mode)

    od_skims = lx.Dataset.construct.from_omx(skims)

    dt = lx.DataTree(
        tour=tour_dataset,
        hh=hh.set_index("HHID"),
        person=pp.set_index("PERSONID"),
        od=od_skims,
        do=od_skims,
        relationships=(
            "tours.HHID @ hh.HHID",
            "tours.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ od.otaz",
            "tours.DTAZ @ od.dtaz",
            "hh.HOMETAZ @ do.dtaz",
            "tours.DTAZ @ do.otaz",
        ),
    )

    dt_work = dt.query_cases("TOURPURP == 1")
    dt_work.n_cases

    dt_work_low_income = dt.query_cases("TOURPURP == 1 and INCOME < 30000")
    dt_work_low_income.n_cases

    m = lx.Model(datatree=dt_work)
    m.title = "Exampville Work Tour Mode Choice v1"

    m.utility_co[Mode.DA] = (
        +P.InVehTime * X.AUTO_TIME + P.Cost * X.AUTO_COST  # dollars per mile
    )

    m.utility_co[Mode.SR] = (
        +P.ASC_SR
        + P.InVehTime * X.AUTO_TIME
        + P.Cost * (X.AUTO_COST * 0.5)  # dollars per mile, half share
        + P("LogIncome:SR") * X("log(INCOME)")
    )

    m.utility_co[Mode.Walk] = (
        +P.ASC_Walk + P.NonMotorTime * X.WALK_TIME + P("LogIncome:Walk") * X("log(INCOME)")
    )

    m.utility_co[Mode.Bike] = (
        +P.ASC_Bike + P.NonMotorTime * X.BIKE_TIME + P("LogIncome:Bike") * X("log(INCOME)")
    )

    m.utility_co[Mode.Transit] = (
        +P.ASC_Transit
        + P.InVehTime * X.TRANSIT_IVTT
        + P.OutVehTime * X.TRANSIT_OVTT
        + P.Cost * X.TRANSIT_FARE
        + P("LogIncome:Transit") * X("log(INCOME)")
    )

    Car = m.graph.new_node(parameter="Mu:Car", children=[Mode.DA, Mode.SR], name="Car")
    NonMotor = m.graph.new_node(
        parameter="Mu:NonMotor", children=[Mode.Walk, Mode.Bike], name="NonMotor"
    )
    Motor = m.graph.new_node(
        parameter="Mu:Motor", children=[Car, Mode.Transit], name="Motor"
    )

    m.graph

    m.choice_co_code = "TOURMODE"

    m.availability_co_vars = {
        Mode.DA: "AGE >= 16",
        Mode.SR: 1,
        Mode.Walk: "WALK_TIME < 60",
        Mode.Bike: "BIKE_TIME < 60",
        Mode.Transit: "TRANSIT_FARE>0",
    }

    m.choice_avail_summary()

    m.compute_engine = "numba"

    # TEST
    # testing the JAX engine
    mj = m.copy()
    mj.compute_engine = "jax"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(method="bhhh")

    m.calculate_parameter_covariance();

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    # testing the JAX engine
    mj.set_cap(20)

    # TEST
    # testing the JAX engine
    resultj = mj.maximize_loglike(stderr=False)

    report = lx.Reporter(title=m.title)

    report.append("# Parameter Summary")
    report.append(m.parameter_summary())
    report

    report << "# Estimation Statistics" << m.estimation_statistics()

    report << "# Utility Functions" << m.utility_functions()

    report.save(
        "exampville_mode_choice.html",
        overwrite=True,
        metadata=m.dumps(),
    )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import xmle.xhtml
from xmle import Elem

from . import styles

# xml.etree.ElementTree.register_namespace("", "http://www.w3.org/2000/svg")
# xml.etree.ElementTree.register_namespace("xlink", "http://www.w3.org/1999/xlink")

# @import url(https://fonts.googleapis.com/css?family=Roboto+Mono:400,700,700italic,400italic,100,100italic);


# 	font-family: "Roboto Slab", Roboto, Helvetica, sans-serif;

_default_css = (
    """

@import url(https://fonts.googleapis.com/css?family=Roboto:400,700,500italic,100italic|Roboto+Mono:300,400,700);

.error_report {color:red; font-family:monospace;}

body {"""
    + styles.body_font
    + """}

div.larch_title {
	font-family: "Book-Antiqua", "Palatino", serif;
	font-size:200%;
	font-weight:900;
	font-style:normal;
	color: #444444;
}

table {border-collapse:collapse;}

table, th, td {
	border: 1px solid #999999;
	font-family:"Roboto Mono", monospace;
	font-size:90%;
	font-weight:400;
	}

th, td { padding:2px; }

td.parameter_category {
	font-family:"Roboto", monospace;
	font-weight:500;
	background-color: #f4f4f4;
	font-style: italic;
	}

th {
	font-family:"Roboto", monospace;
	font-weight:700;
	}

.larch_signature {"""
    + styles.signature_font
    + """}
.larch_name_signature {"""
    + styles.signature_name_font
    + """}

a.parameter_reference {font-style: italic; text-decoration: none}

.strut2 {min-width:1in}

.histogram_cell { padding-top:1; padding-bottom:1; vertical-align:center; }

.dicta pre {
	margin:0;
	font-family:"Roboto Mono", monospace;
	font-weight:300;
	font-size:70%;
}

.raw_log pre {
	font-family:"Roboto Mono", monospace;
	font-weight:300;
	font-size:70%;
	}

caption {
    caption-side: bottom;
	text-align: left;
	font-family: Roboto;
	font-style: italic;
	font-weight: 100;
	font-size: 80%;
}

table.dictionary { border:0px hidden !important; border-collapse: collapse !important; }
div.blurb {
	margin-top: 15px;
	max-width: 6.5in;
}

div.note {
	font-size:90%;
	padding-left:1em;
	padding-right:1em;
	border: 1px solid #999999;
	border-radius: 4px;
}

p.admonition-title {
	font-weight: 700;
}

.tooltipped {
	position: relative;
	display: inline-block;
}

.tooltipped .tooltiptext {
	visibility: hidden;
	width: 180px;
	background-color: black;
	color: #fff;
	text-align: center;
	border-radius: 6px;
	padding: 5px 0;
	position: absolute;
	z-index: 1;
	top: -5px;
	left: 110%;
}

.tooltipped .tooltiptext::after {
	content: "";
	position: absolute;
	top: 50%;
	right: 100%;
	margin-top: -5px;
	border-width: 5px;
	border-style: solid;
	border-color: transparent black transparent transparent;
}
.tooltipped:hover .tooltiptext {
	visibility: visible;
}

"""
)


def larch_style():
    return Elem(
        tag="style",
        text=_default_css,
    )


_tooltipped_style_css = """
		.tooltipped {
			position: relative;
			display: inline-block;
		}

		.tooltipped .tooltiptext {
			visibility: hidden;
			width: 180px;
			background-color: black;
			color: #fff;
			text-align: center;
			border-radius: 6px;
			padding: 5px 0;
			position: absolute;
			z-index: 1;
			top: -5px;
			left: 110%;
		}

		.tooltipped .tooltiptext::after {
			content: "";
			position: absolute;
			top: 50%;
			right: 100%;
			margin-top: -5px;
			border-width: 5px;
			border-style: solid;
			border-color: transparent black transparent transparent;
		}
		.tooltipped:hover .tooltiptext {
			visibility: visible;
		}




	"""


def tooltipped_style():
    return Elem("style", attrib={"type": "text/css"}, text=_tooltipped_style_css)


def floating_table_head():
    floatThead = Elem(
        tag="script",
        attrib={
            "src": "https://cdnjs.cloudflare.com/ajax/libs/floatthead/2.0.3/jquery.floatThead.min.js",
        },
    )
    floatTheadA = Elem(
        tag="script",
        text="""
	$( document ).ready(function() {
		var $table = $('table.floatinghead');
		$table.floatThead({ position: 'absolute' });
		var $tabledf = $('table.dataframe');
		$tabledf.floatThead({ position: 'absolute' });
	});
	$(window).on("hashchange", function () {
		window.scrollTo(window.scrollX, window.scrollY - 50);
	});
	""",
    )
    return floatThead, floatTheadA


def logo_in_signature():
    from .. import __version__ as version
    from .images import favicon

    sig = Elem("div")
    sig << Elem(
        "img",
        {
            "width": "14",
            "height": "14",
            "src": f"data:image/png;base64,{favicon}",
            "style": "position:relative;top:2px;",
        },
    )
    sig << Elem("span", {"class": "larch_name_signature"}, text=f" Larch {version}")
    return sig


xmle.xhtml.logo_in_signature = logo_in_signature



def example(extract='m', estimate=False):
    # TEST
    import numpy as np
    import pandas as pd

    pd.set_option("display.max_columns", 999)
    pd.set_option("expand_frame_repr", False)
    pd.set_option("display.precision", 3)
    np.set_printoptions(precision=12)

    from pytest import approx

    import larch as lx

    import numpy as np
    import pandas as pd

    import larch as lx
    from larch import P, X

    print(lx.versions())

    raw = pd.read_csv(lx.example_file("swissmetro.csv.gz"))
    raw["SM_COST"] = raw["SM_CO"] * (raw["GA"] == 0)
    raw["TRAIN_COST"] = raw.eval("TRAIN_CO * (GA == 0)")
    raw["TRAIN_COST_SCALED"] = raw["TRAIN_COST"] / 100
    raw["TRAIN_TT_SCALED"] = raw["TRAIN_TT"] / 100
    raw["SM_COST_SCALED"] = raw.eval("SM_COST / 100")
    raw["SM_TT_SCALED"] = raw["SM_TT"] / 100
    raw["CAR_CO_SCALED"] = raw["CAR_CO"] / 100
    raw["CAR_TT_SCALED"] = raw["CAR_TT"] / 100
    raw["CAR_AV_SP"] = raw.eval("CAR_AV * (SP!=0)")
    raw["TRAIN_AV_SP"] = raw.eval("TRAIN_AV * (SP!=0)")

    raw

    data = lx.Dataset.construct.from_idco(raw).dc.query_cases(
        "PURPOSE in (1,3) and CHOICE != 0"
    )

    data

    availability_co_vars = {
        1: "TRAIN_AV_SP",
        2: "SM_AV",
        3: "CAR_AV_SP",
    }

    m1 = lx.Model(title="Model1")
    m1.availability_co_vars = availability_co_vars
    m1.choice_co_code = "CHOICE"
    m1.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_COST_SCALED") * P("B_COST")
    m1.utility_co[2] = X("SM_COST_SCALED") * P("B_COST")
    m1.utility_co[3] = P("ASC_CAR") + X("CAR_CO_SCALED") * P("B_COST")

    m2 = lx.Model(title="Model2")
    m2.availability_co_vars = availability_co_vars
    m2.choice_co_code = "CHOICE"
    m2.utility_co[1] = (
        P("ASC_TRAIN")
        + X("TRAIN_TT_SCALED") * P("B_TIME")
        + X("TRAIN_COST_SCALED") * P("B_COST")
    )
    m2.utility_co[2] = X("SM_TT_SCALED") * P("B_TIME") + X("SM_COST_SCALED") * P("B_COST")
    m2.utility_co[3] = (
        P("ASC_CAR") + X("CAR_TT_SCALED") * P("B_TIME") + X("CAR_CO_SCALED") * P("B_COST")
    )

    m3 = lx.Model(title="Model3")
    m3.availability_co_vars = availability_co_vars
    m3.choice_co_code = "CHOICE"
    m3.utility_co[1] = X("TRAIN_COST_SCALED") * P("Z_COST")
    m3.utility_co[2] = X("SM_COST_SCALED") * P("Z_COST")
    m3.utility_co[3] = X("CAR_CO_SCALED") * P("Z_COST")

    mk = lx.Model()
    mk.utility_co[102] = P("W_OTHER")
    mk.utility_co[103] = P("W_COST")

    b = lx.LatentClass(
        mk,
        {101: m1, 102: m2, 103: m3},
        datatree=data.dc.set_altids([1, 2, 3]),
        groupid="ID",
    )

    b.lock_value(Z_COST=-10000)

    b.set_cap(25)
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = b.maximize_loglike(method="slsqp")

    b.calculate_parameter_covariance();

    b.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import datetime
import pprint

from .addict_yaml import Dict as Dict
from .data_expansion import hard_sigmoid as hard_sigmoid
from .data_expansion import piece as piece
from .signal_dict import SignalDict as SignalDict


def _prettyprint_fallback(x):
    if isinstance(x, dictx):
        return repr(x)
    else:
        return pprint.pformat(x)


class dictx(dict):
    """Python dict with attribute access and xml output."""

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(str(_) for _ in self.keys()))) + 1
            return "\n".join(
                [
                    "┣"
                    + str(k).rjust(m)
                    + ": "
                    + _prettyprint_fallback(v).replace("\n", "\n┃" + " " * (m + 2))
                    for k, v in self.items()
                ]
            )
        else:
            return self.__class__.__name__ + "()"

    def __getattr__(self, item):
        if item[0] == "_" and item[-1] == "_":
            raise AttributeError(item)
        return self.__getitem__(item)

    def __setattr__(self, name, value):
        if hasattr(dictx, name):
            raise AttributeError(f"'dictx' object attribute '{name}' is read-only")
        else:
            self[name] = value

    def __xml__(self):
        from xmle import Elem, Show

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;")
        if len(self):
            tr = t.elem("tr")
            tr.elem("th", text="key")
            tr.elem("th", text="value", style="text-align:left;")
            for k, v in self.items():
                tr = t.elem("tr")
                tr.elem("td", text=str(k))
                try:
                    if isinstance(v, str) and "\n" not in v and v[:1] != "#":
                        raise ValueError()
                    v_ = Show(v)
                except (AttributeError, ValueError):
                    plaintext = pprint.pformat(v)
                    if isinstance(v, datetime.timedelta):
                        plaintext = str(v)
                    if "\n" in plaintext:
                        tr.elem("td", style="text-align:left;").elem(
                            "pre", text=plaintext
                        )
                    else:
                        tr.elem("td", style="text-align:left;", text=plaintext)
                else:
                    tr.elem("td", style="text-align:left;") << v_
        else:
            tr = t.elem("tr")
            tr.elem("td", text="<empty>")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def copy(self):
        return dictx(super().copy())


dicta = Dict


# Add statistics to pandas.DataFrame and pandas.Series
# import pandas
# from .statistics import statistics_for_dataframe, statistics_for_array5, uniques, invmap
# pandas.DataFrame.statistics = statistics_for_dataframe
# pandas.Series.statistics = statistics_for_array5
# pandas.Series.uniques = uniques
# pandas.Series.invmap = invmap
#
# from .dataframe import compute
# pandas.DataFrame.compute = compute



from __future__ import annotations

import logging
from typing import Literal

import numpy as np
import pandas as pd
import xarray as xr

from ..util import dictx
from .basemodel import BaseModel
from .numbamodel import NumbaModel as Model
from .possible_overspec import (
    PossibleOverspecificationError,
    compute_possible_overspecification,
)

logger = logging.getLogger(__name__)


def doctor(
    model: Model,
    *,
    repair_ch_av: Literal["?", "+", "-", "!"] | None = "?",
    repair_ch_zq: Literal["?", "-", "!"] | None = None,
    repair_av_zq: Literal["?", "-", "!"] | None = None,
    repair_noch_nzwt: Literal["?", "+", "-"] | None = "?",
    repair_nan_wt: Literal["?", True, "!"] | None = "?",
    repair_nan_data_co: Literal["?", True, "!"] | None = "?",
    check_low_variance_data_co: Literal["?", "!"] | None = None,
    check_overspec: Literal["?", "!", None] = None,
    repair_nan_utility: Literal["?", True, "!"] | None = "?",
    verbose: int = 3,
    warning_stacklevel: int = 2,
):
    """
    Diagnose data problems with a model.

    The doctor will check for common data problems that can cause numerical
    instability in a model.  The doctor will return a list of problems found,
    and optionally repair them.

    Parameters
    ----------
    model : larch.Model
        The model to diagnose.
    repair_ch_av : {'?', '+', '-', '!'}, default '?'
        How to repair the data if some observations are chosen but not available.
        The plus ('+') will make the conflicting alternatives available, overriding
        the availability status. The minus ('-') will make them not chosen (possibly
        leaving no chosen alternative). A question mark ('?') effects no repair, and
        simply emits a warning without interrupting program execution. An exclamation
        mark will raise an error if there are any conflicts.
    repair_ch_zq : {'?', '-', '!'}, default None
        How to repair the data if some observations are chosen but have zero quantity.
        The minus ('-') will make alternatives with zero quantity not chosen (possibly
        leaving no chosen alternative). A question mark ('?') effects no repair, and
        simply emits a warning. An exclamation mark ('!') will raise an error if there
        are any conflicts.
    repair_av_zq : {'?', '-', '!'}, default None
        How to repair the data if some observations are available but have zero quantity.
        The minus ('-') will make alternatives with zero quantity not available (possibly
        leaving no available alternatives). A question mark ('?') effects no repair, and
        simply emits a warning. An exclamation mark ('!') will raise an error if there are
        any conflicts.
    repair_noch_nzwt : {'?', '+', '-'}, default '?'
        How to repair the data if some observations have no choice but have some weight.
        Minus ('-') will make the weight zero when there is no choice. Plus ('+') will
        make the weight zero, plus autoscale all remaining weights so the total of the
        case weights equals the number of cases. A question mark ('?') effects no repair,
        and simply emits a warning.
    repair_nan_wt : {'?', '!', True}, default '?'
        How to repair the data if some weight values are NaN. Any true value other than
        "?" or "!" will make NaN values in weight zero. The question mark simply emits
        a warning if there are NaN values found, while the exclamation mark will raise
        an error.
    repair_nan_data_co : {'?', '!', True}, default '?'
        How to repair the data if some data_co values are NaN. Any true value other than
        "?" or "!" will make NaN values in data_co zero. The question mark simply emits
        a warning if there are NaN values found, while the exclamation mark will raise
        an error.
    check_low_variance_data_co : {'?', '!'}, default None
        Check if any data_co columns have very low variance. No repairs are available for
        this check. The question mark simply emits a warning if there are issues found,
        while the exclamation mark will raise an error.
    check_overspec : {'?', '!'}, default None
        Check model for possible over-specification. No repairs are available for this
        check. A question mark ('?') simply emits a warning if a possible over-
        specification is found. An exclamation mark ('!') will raise an error if
        possible over-specification is found.  This is considered a "deep" check, and
        will only be run if there are no known data problems found by the other checks.
    repair_nan_utility : {'?', '!', True}, default '?'
        How to repair the data if some utility values are NaN at current parameters.
        Any true value other than "?" or "!" will take alternatives with NaN values in
        utility, and make them unavailable. The question mark simply emits a warning if
        there are NaN values found, while the exclamation mark will raise an error. This
        is considered a "deep" check, and will only be run if there are no known data
        problems found by the other checks.
    verbose : int, default 3
        The number of example rows to list for each problem.
    warning_stacklevel : int, default 2
        The stacklevel for warnings.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    problems : dict
        A dictionary of problems found, with the key being the name of the problem
        and the value being a DataFrame with the number of bad instances and some
        example rows.

    Raises
    ------
    TypeError
        If the model is not a Model instance.
    ValueError
        If any of the repair settings are invalid, or if the repair is set to '!' and
        there are any conflicts found.
    """
    problems = dictx()

    if not isinstance(model, Model):
        raise TypeError("the doctor requires a Model instance to diagnose")

    def apply_repair(repair, repair_func):
        nonlocal model, problems, verbose
        if repair is None:
            return
        logger.info(f"checking for {repair_func.__name__}")
        model, diagnosis = repair_func(model, repair=repair, verbose=verbose)
        if diagnosis is not None:
            logger.warning(
                f"problem: {repair_func.__name__} has ({len(diagnosis)} issues)",
                stacklevel=warning_stacklevel + 2,
            )
            problems[repair_func.__name__] = diagnosis

    apply_repair(repair_ch_av, chosen_but_not_available)
    apply_repair(repair_noch_nzwt, nothing_chosen_but_nonzero_weight)
    apply_repair(repair_nan_data_co, nan_data_co)
    apply_repair(repair_nan_wt, nan_weight)
    apply_repair(check_low_variance_data_co, low_variance_data_co)
    apply_repair(repair_ch_zq, chosen_but_zero_quantity)
    apply_repair(repair_av_zq, available_but_zero_quantity)

    if not problems:
        # Following are deep checks, which require actually evaluating the model.
        # We run these after the data checks above, so we can skip this work if there
        # are already known data problems.
        apply_repair(repair_nan_utility, nan_utility)
        apply_repair(check_overspec, overspecification)

    return model, problems


def chosen_but_not_available(
    model: Model, repair: Literal["?", "+", "-", "!"] = "?", verbose: int = 3
) -> tuple[Model, pd.DataFrame | None]:
    """
    Check if some observations are chosen but not available.

    Alternatives that are unavailable have their utility values set to negative
    infinity.  If even one observation is chosen but not available, the model
    log-likelihood will nominally be negative infinity regardless of the values
    of any other parameters.  Note that some compute engines (e.g. JAX) may
    not actually return negative infinity log likelihoods due to clipping of
    extreme values.

    Parameters
    ----------
    model : larch.Model
        The model to check.
    repair : {'?', '+', '-', '!'}, default '?'
        How to repair the data. The plus ('+') will make the conflicting
        alternatives available, overriding the availability status. The minus
        ('-') will make them not chosen (possibly leaving no chosen alternative).
        A question mark ('?') effects no repair, and simply emits a warning
        without interrupting program execution. An exclamation mark will raise
        an error if there are any conflicts.
    verbose : int, default 3
        The number of example rows to list for each problem.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    diagnosis : pd.DataFrame
        The number of bad instances, by alternative, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any conflicts found.
    """
    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    not_avail = dataset["av"] == 0
    chosen = dataset["ch"] > 0
    chosen_but_not_available = not_avail & chosen
    chosen_but_not_available_sum = chosen_but_not_available.sum(dataset.dc.CASEID)

    diagnosis = None
    if chosen_but_not_available_sum.sum() > 0:
        i1, i2 = np.where(chosen_but_not_available)

        diagnosis = pd.DataFrame(
            chosen_but_not_available_sum[chosen_but_not_available_sum > 0],
            columns=[
                "n",
            ],
        )
        diagnosis.insert(0, "altid", None)

        diagnosis_rownum = 0
        for colnum, colname in enumerate(
            chosen_but_not_available.coords[dataset.dc.ALTID]
        ):
            if chosen_but_not_available_sum[colnum] > 0:
                diagnosis.loc[diagnosis_rownum, "example rows"] = ", ".join(
                    str(j) for j in i1[i2 == colnum][:verbose]
                )
                if isinstance(colname, xr.DataArray) and colname.ndim == 0:
                    colname = colname.item()
                diagnosis.loc[diagnosis_rownum, "altid"] = colname
                diagnosis_rownum += 1

        if repair == "+":
            model.dataset["av"].data[
                chosen_but_not_available.values[:, : model.dataset["av"].shape[1]]
            ] = 1
        elif repair == "-":
            model.dataset["ch"].data[
                chosen_but_not_available.values[:, : model.dataset["ch"].shape[1]]
            ] = 0
        elif repair == "!":
            raise ValueError(
                "some observed choices are not available (try `repair_ch_av`)"
            )

    return model, diagnosis


def chosen_but_zero_quantity(
    model: Model, repair: Literal["?", "-", "!"] | None = None, verbose: int = 3
):
    """
    Check if some observations are chosen but have zero quantity.

    Alternatives that have zero quantity have utility values that end up as
    negative infinity, regardless of whether the alternative would otherwise be
    available.  Due to the mathematical structure of how quantities are used in
    Larch, this situation is generally a result of a data problem, and not
    due to the current values of model parameters.

    If even one observation is chosen but has zero quantity, the model
    log-likelihood will nominally be negative infinity regardless of the values
    of any other parameters.  Note that some compute engines (e.g. JAX) may
    not actually return negative infinity log likelihoods due to clipping of
    extreme values.

    Parameters
    ----------
    model : BaseModel
        The model to check.
    repair : {'?', '-', '!'}
        How to repair the data. The minus ('-') will make alternatives with zero
        quantity not chosen (possibly leaving no chosen alternative). A question
        mark ('?') effects no repair, and simply emits a warning. An exclamation
        mark ('!') will raise an error if there are any conflicts.
    verbose : int, default 3
        The number of example rows to list for each problem.

    Returns
    -------
    model : BaseModel
        The model with revised data
    diagnosis : pd.DataFrame
        The number of bad instances, by alternative, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any conflicts found.
    """
    if repair not in ("?", "-", "!", None):
        raise ValueError(f'invalid repair setting "{repair}"')

    if not model.quantity_ca:
        # no quantities, so no problem
        return model, None

    quant = model.quantity()
    zero_quantity = np.asarray(quant[:, : model.graph.n_elementals()] == 0)

    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    ch_and_zero_quantity = (dataset["ch"] > 0) & zero_quantity
    ch_and_zero_quantity_sum = ch_and_zero_quantity.sum(dataset.dc.CASEID)
    diagnosis = None
    if ch_and_zero_quantity.sum() > 0:
        i1, i2 = np.where(ch_and_zero_quantity)
        diagnosis = (
            ch_and_zero_quantity_sum[ch_and_zero_quantity_sum > 0]
            .to_pandas()
            .rename("n")
            .to_frame()
        )

        for colnum, colname in enumerate(dataset.dc.altids()):
            if ch_and_zero_quantity_sum[colnum] > 0:
                diagnosis.loc[colname, "example rows"] = ", ".join(
                    str(j) for j in i1[i2 == colnum][:verbose]
                )

        msg = "chosen_but_zero_quantity: some observed choices have zero quantity.\n"
        try:
            from tabulate import tabulate
        except ImportError:
            msg += diagnosis.to_string()
        else:
            msg += tabulate(diagnosis, headers="keys", tablefmt="fancy_outline")
        if repair == "!":
            msg += "\nTry `repair_ch_zq` to resolve."
            raise ValueError(msg)
        elif repair == "?":
            logger.warning(msg)
        elif repair == "-":
            logger.warning(
                msg.replace(
                    "some observed choices", "zeroing out observed choices that"
                )
            )
            model.dataset["ch"].values[ch_and_zero_quantity] = 0

    return model, diagnosis


def available_but_zero_quantity(
    model: Model, repair: Literal["?", "-", "!"] | None = None, verbose: int = 3
):
    """
    Check if some observations are available but have zero quantity.

    Alternatives that have zero quantity have utility values that end up as
    negative infinity, regardless of whether the alternative would otherwise be
    available.  Due to the mathematical structure of how quantities are used in
    Larch, this situation is generally a result of a data problem, and not
    due to the current values of model parameters.

    If even one observation is available but has zero quantity, the first
    derivative of the model log-likelihood may be incalculable, and the model
    parameter estimation process may fail.

    Parameters
    ----------
    model : BaseModel
        The model to check.
    repair : {'?', '-', '!'}
        How to repair the data. The minus ('-') will make alternatives with zero
        quantity not available (possibly leaving no available alternatives). A
        question mark ('?') effects no repair, and simply emits a warning. An
        exclamation mark ('!') will raise an error if there are any conflicts.
    verbose : int, default 3
        The number of example rows to list for each problem.

    Returns
    -------
    model : BaseModel
        The model with revised data
    diagnosis : pd.DataFrame
        The number of bad instances, by alternative, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any conflicts found.
    """
    if repair not in ("?", "-", "!", None):
        raise ValueError(f'invalid repair setting "{repair}"')

    if not model.quantity_ca:
        # no quantities, so no problem
        return model, None

    quant = model.quantity()
    zero_quantity = np.asarray(quant[:, : model.graph.n_elementals()] == 0)

    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    av_and_zero_quantity = (dataset["av"] > 0) & zero_quantity
    av_and_zero_quantity_sum = av_and_zero_quantity.sum(dataset.dc.CASEID)
    diagnosis = None
    if av_and_zero_quantity.sum() > 0:
        i1, i2 = np.where(av_and_zero_quantity)
        diagnosis = (
            av_and_zero_quantity_sum[av_and_zero_quantity_sum > 0]
            .to_pandas()
            .rename("n")
            .to_frame()
        )

        for colnum, colname in enumerate(dataset.dc.altids()):
            if av_and_zero_quantity_sum[colnum] > 0:
                diagnosis.loc[colname, "example rows"] = ", ".join(
                    str(j) for j in i1[i2 == colnum][:verbose]
                )

        msg = (
            "available_but_zero_quantity: some available choices have zero quantity.\n"
        )
        try:
            from tabulate import tabulate
        except ImportError:
            msg += diagnosis.to_string()
        else:
            msg += tabulate(diagnosis, headers="keys", tablefmt="fancy_outline")
        if repair == "!":
            msg += "\nTry `repair_av_zq` to resolve."
            raise ValueError(msg)
        elif repair == "?":
            logger.warning(msg)
        elif repair == "-":
            logger.warning(
                msg.replace(
                    "some available choices", "zeroing out available choices that"
                )
            )
            model.dataset["av"].values[av_and_zero_quantity] = 0

    return model, diagnosis


def nothing_chosen_but_nonzero_weight(
    model, repair: Literal["?", "-", "*", "!"] = "?", verbose=3
):
    """
    Check if some observations have no choice but have some weight.

    Parameters
    ----------
    model : BaseModel
        The model to check.
    repair : {'?', '-', '*', '!'}
        How to repair the data. Minus ('-') will make the weight zero when there
        is no choice. Star ('*') will also make the weight zero, plus autoscale
        all remaining weights so the total of the case weights equals the number
        of cases. A question mark ('?') effects no repair, and simply emits a
        warning.
    verbose : int, default 3
        The number of example rows to list for each problem.

    Returns
    -------
    model : BaseModel
        The revised dataframe
    diagnosis : pd.DataFrame
        The number of bad instances, by alternative, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any conflicts found.
    """
    diagnosis = None
    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    if "wt" in dataset and "ch" in dataset:
        nothing_chosen = dataset["ch"].sum(dataset.dc.ALTID) == 0
        nothing_chosen_some_weight = nothing_chosen & (dataset["wt"] > 0)
        if nothing_chosen_some_weight.sum() > 0:
            i1 = np.where(nothing_chosen_some_weight)[0]

            diagnosis = pd.DataFrame(
                [
                    nothing_chosen_some_weight.sum(),
                ],
                columns=[
                    "n",
                ],
                index=[
                    "nothing_chosen_some_weight",
                ],
            )

            diagnosis.loc["nothing_chosen_some_weight", "example rows"] = ", ".join(
                str(j) for j in i1[:verbose]
            )
            msg = "nothing_chosen_but_nonzero_weight: some cases have no choice but non-zero weight.\n"
            try:
                from tabulate import tabulate
            except ImportError:
                msg += diagnosis.to_string()
            else:
                msg += tabulate(diagnosis, headers="keys", tablefmt="fancy_outline")
            if repair == "!":
                msg += "\nTry `repair_noch_nzwt` to resolve."
                raise ValueError(msg)
            elif repair == "?":
                logger.warning(msg)
            elif repair == "+":
                raise ValueError(
                    "cannot resolve nothing_chosen_but_nonzero_weight by assuming some choice"
                )
            elif repair == "-":
                model.dataset["wt"].data[nothing_chosen] = 0
            elif repair == "*":
                model.dataset["wt"].data[nothing_chosen] = 0
                model.dataset.dc.autoscale_weights()
    return model, diagnosis


def nan_data_co(
    model: Model, repair: Literal["?", True, "!"] = "?", verbose: int = 3
) -> tuple[Model, pd.DataFrame | None]:
    """
    Check if some data_co values are NaN.

    Parameters
    ----------
    model : larch.Model
        The model to check.
    repair : {"?", "!", True}
        Whether to repair the data. Any true value other than "?" or "!" will make
        NaN values in data_co zero. The question mark simply emits a warning
        if there are NaN values found, while the exclamation mark will raise an error.
    verbose : int, default 3
        The number of example columns to list for each problem.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    diagnosis : pd.DataFrame
        The number of bad instances, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any NaN values found.
    """
    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    diagnosis = None
    if "co" in dataset:
        nan_dat = np.isnan(dataset["co"]).sum(dataset.dc.CASEID)
        if nan_dat.sum():
            diagnosis = (
                nan_dat[nan_dat > 0]
                .iloc[:verbose]
                .to_pandas()
                .rename("n_nan")
                .to_frame()
            )
            n = int(nan_dat.sum())
            if repair == "?":
                logger.warning(f"nan_data_co: {n} instances have NaN values")
            elif repair == "!":
                raise ValueError(
                    f"nan_data_co: {n} instances have NaN values, try `repair_nan_data_co`"
                )
            if repair and repair != "?":
                dataset["co"] = dataset["co"].fillna(0)

    return model, diagnosis


def nan_weight(
    model: Model, repair: Literal["?", True, "!"] = "?", verbose: int = 3
) -> tuple[Model, pd.DataFrame | None]:
    """
    Check if some weight values are NaN.

    Parameters
    ----------
    model : larch.Model
        The model to check.
    repair : {"?", "!", True}
        Whether to repair the data. Any true value other than "?" or "!" will make
        NaN values in weight zero. The question mark simply emits a warning
        if there are NaN values found, while the exclamation mark will raise an error.
    verbose : int, default 3
        The number of example columns to list for each problem.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    diagnosis : pd.DataFrame
        The number of bad instances, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any NaN values found.
    """
    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    diagnosis = None
    if "wt" in dataset:
        nan_wt = int(np.isnan(dataset["wt"]).sum())
        if nan_wt:
            diagnosis = (
                dataset["wt"][np.isnan(dataset["wt"])]
                .iloc[:verbose]
                .to_pandas()
                .to_frame()
            )
            if repair == "?":
                logger.warning(f"nan_weight: {nan_wt} instances have NaN values")
            elif repair == "!":
                raise ValueError(
                    f"nan_weight: {nan_wt} instances have NaN values, try `repair_nan_wt`"
                )
            if repair and repair != "?":
                dataset["wt"] = dataset["wt"].fillna(0)

    return model, diagnosis


def low_variance_data_co(
    model: Model, repair: Literal["?", "!"] = "?", verbose: int = 3
):
    """
    Check if any data_co columns have very low variance.

    Parameters
    ----------
    model : larch.Model
        The model to check.
    repair : {"?", "!"}
        No repairs are available for this check. The question mark simply emits
        a warning if there are issues found, while the exclamation mark will
        raise an error.
    verbose : int, default 3
        The number of example columns to list if there is a problem.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    diagnosis : pd.DataFrame
        The number of bad instances, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any low variance columns found.
    """
    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    diagnosis = None
    if "co" in dataset:
        variance = dataset["co"].var(dataset.dc.CASEID).to_pandas().rename("variance")
        if variance.min() < 1e-3:
            diagnosis = variance[variance < 1e-3].to_frame()
            if repair == "?":
                logger.warning(
                    f"low_variance_data_co: {len(diagnosis)} columns have low variance"
                )
            elif repair == "!":
                raise ValueError(
                    f"low_variance_data_co: {len(diagnosis)} columns have low variance"
                )

    return model, diagnosis


def nan_utility(model: Model, repair: Literal["?", True, "!"] = "?", verbose: int = 3):
    """
    Check if any utility values are NaN at current parameters.

    Parameters
    ----------
    model : larch.Model
        The model to check.
    repair : {'?', '!', True}
        Whether to repair the data. Any true value other than "?" or "!" will
        take alternatives with NaN values in utility, and make them unavailable.
        The question mark simply emits a warning if there are NaN values found,
        while the exclamation mark will raise an error.
    verbose : int, default 3
        The number of example columns to list for each problem.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    diagnosis : pd.DataFrame
        The number of bad instances, and some example rows.

    Raises
    ------
    ValueError
        If the repair is set to '!' and there are any NaN values found.
    """
    dataset = model.dataset
    if dataset is None:
        raise ValueError("data not loaded")
    assert isinstance(dataset, xr.Dataset)

    n_alts = dataset["av"].shape[1]

    u = model.utility()[:, :n_alts]
    nan_u = np.isnan(u) & (dataset["av"] > 0)

    diagnosis = None
    nan_util = int(nan_u.sum())
    if nan_util:
        diagnosis = nan_u.sum(dataset.dc.CASEID).to_pandas().rename("n").to_frame()
        if repair == "?":
            logger.warning(
                f"nan_utility: {nan_util} available alternatives have NaN "
                f"utility values"
            )
        elif repair == "!":
            raise ValueError(
                f"nan_utility: {nan_util} available alternatives have NaN "
                f"utility values,\ntry using the Model.doctor `repair_nan_utility` "
                f"argument, set it to True to make them unavailable"
            )
        elif repair:
            model.dataset["av"].data[nan_u] = 0

    return model, diagnosis


def overspecification(
    model: BaseModel, repair: Literal["?", "!"] = "?", verbose: int = 3
):
    """
    Check model for possible over-specification.

    Parameters
    ----------
    model : larch.Model
        The model to check.
    repair : {'?', '!'}
        No automatic repairs are available for this check. A question mark ('?') simply
        emits a warning if a possible over-specification is found. An exclamation mark
        ('!') will raise an error if possible over-specification is found.
    verbose : int, default 3
        This is ignored for the overspecification check; all possible problems are
        listed.

    Returns
    -------
    model : larch.Model
        The model with revised dataset attached.
    diagnosis : pd.DataFrame
        A dataframe of possible over-specification problems in the model.  The index of
        this dataframe is a multi-index with the first level being the problem number, the
        second level being the eigenvalue, and the third level being the parameter name[s]
        of the non-zero elements of each problematic eigenvector. The columns are the
        non-zero eigenvector values.
    """
    pvals = model.pvals
    locks = np.asarray(model.pholdfast.astype(bool))
    if model.compute_engine == "jax":
        _se, hess, _inv_hess = model.jax_param_cov(pvals)
    else:
        hess = -model.d2_loglike(pvals)
    hess = np.asarray(hess).copy()
    hess[locks, :] = 0
    hess[:, locks] = 0

    diagnosis = None
    overspec = compute_possible_overspecification(hess, model.pholdfast)
    if overspec:
        diagnosis = []
        possible_overspecification = []
        msg = "Model is possibly over-specified (hessian is nearly singular)."
        msg += "\nLook for problems in these parameters or groups of parameters:"
        for eigval, ox, eigenvec in overspec:
            if eigval == "LinAlgError":
                possible_overspecification.append((eigval, [ox], [""]))
            else:
                paramset = list(np.asarray(model.pnames)[ox])
                possible_overspecification.append((eigval, paramset, eigenvec[ox]))
                diagnosis.append(
                    (
                        eigval,
                        pd.Series(eigenvec[ox], index=paramset, name="eigenvector"),
                    )
                )
                msg += f"\n- Eigenvalue: {eigval}"
                max_len_param = max(len(p) for p in paramset)
                for p, z in zip(paramset, eigenvec[ox]):
                    msg += f"\n    {p:{max_len_param}s}: {z}"
        model._possible_overspecification = possible_overspecification
        if repair == "!":
            raise PossibleOverspecificationError(msg)
        elif repair == "?":
            logger.warning(msg)
        diagnosis = pd.concat(
            [d[1] for d in diagnosis],
            keys=[(n, d[0]) for (n, d) in enumerate(diagnosis)],
            names=["problem", "eigenvalue", "parameter"],
        )
        if isinstance(diagnosis, pd.Series):
            diagnosis = diagnosis.to_frame()
    return model, diagnosis



# Dense
from __future__ import annotations

CASEID = "_caseid_"
ALTID = "_altid_"

# Sparse
CASEALT = "_casealt_"
ALTIDX = "_alt_idx_"
CASEPTR = "_caseptr_"
GROUPID = "_groupid_"
INGROUP = "_ingroup_"



def example(extract='m', estimate=False):
    import larch as lx

    lx.__version__

    d = lx.examples.MTC(format="dataset")
    d

    m = lx.Model(d, compute_engine="numba")

    from larch import PX, P, X

    for a in [2, 3]:
        m.utility_co[a] = +P("hhinc#2,3") * X("hhinc") + P(f"numveh#{a}") * X("numveh")

    for a in [4, 5, 6]:
        m.utility_co[a] = +P(f"hhinc#{a}") * X("hhinc") + P(f"numveh#{a}") * X("numveh")

    d.dc.alts_mapping

    for a, name in d.dc.alts_mapping.items():
        if a == 1:
            continue
        m.utility_co[a] += P("ASC_" + name)

    m.utility_ca = (
        +PX("totcost")
        + P("motorized_time") * X("(altid <= 4) * tottime")
        + P("nonmotorized_time") * X("(altid > 4) * tottime")
        + P("motorized_ovtbydist") * X("(altid <= 4) * ovtt/dist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 10, Autos per Household"

    m.choice_avail_summary()

    m.set_cap(20)

    # TEST
    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, method="bhhh")

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        ("Ownership", "numveh.*"),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    from pytest import approx

    assert result.loglike == approx(-3501.6427674451616)
    revealed_x = dict(zip(m.pnames, result.x))
    expected_x = {
        "ASC_Bike": -2.220318494507017,
        "ASC_SR2": -2.0539566005386884,
        "ASC_SR3+": -3.643068905518874,
        "ASC_Transit": 0.5738157734892145,
        "ASC_Walk": -0.4403348747865274,
        "hhinc#2,3": -0.001955401039169736,
        "hhinc#4": -0.0013233467107319043,
        "hhinc#5": -0.009513056978703192,
        "hhinc#6": -0.004156633998143166,
        "motorized_ovtbydist": -0.17850538190676227,
        "motorized_time": -0.03782098120360213,
        "nonmotorized_time": -0.047506740267090244,
        "numveh#2": -0.035252140629887845,
        "numveh#3": 0.07226723634772537,
        "numveh#4": -0.5544590047253285,
        "numveh#5": -0.22920682058782335,
        "numveh#6": -0.36557711952920985,
        "totcost": -0.004059016748597727,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(expected_x[k], 2e-2), (
            f"{k}, {revealed_x[k] / expected_x[k]}"
        )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import jax  # noqa: F401
import numpy as np
import pandas as pd
from pytest import approx, fixture

import larch as lx
from larch import PX, P, X


@fixture(scope="module")
def raw_data() -> pd.DataFrame:
    return pd.read_parquet(lx.example_file("vehicle_choice.parquet"))


@fixture
def simple_model(raw_data: pd.DataFrame) -> lx.Model:
    data = lx.Dataset.construct.from_idca(raw_data)
    simple = lx.Model(data)
    # price and opcost are scaled to improve numerical stability
    simple.utility_ca = (
        P("price") * X("price / 10000")
        + P("opcost") * X("opcost / 10")
        + PX("max_range")
        + PX("ev")
        + PX("hybrid")
        + PX("hiperf")
        + PX("medhiperf")
    )
    simple.choice_ca_var = "chosen"
    return simple


def test_vehicle_choice(simple_model: lx.Model):
    simple = simple_model
    simple.maximize_loglike(stderr=True, options={"ftol": 1e-9})
    assert simple.most_recent_estimation_result["loglike"] == approx(-1399.1932)
    expected_value = {
        "price": -0.4167,
        "opcost": -0.1288,
        "max_range": 0.4770,
        "ev": -1.3924,
        "hybrid": 0.3555,
        "hiperf": 0.1099,
        "medhiperf": 0.3841,
    }
    assert simple.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=1e-3
    )
    expected_stderr = {
        "price": 0.0332,
        "opcost": 0.0353,
        "max_range": 0.1765,
        "ev": 0.2766,
        "hybrid": 0.1218,
        "hiperf": 0.0838,
        "medhiperf": 0.0855,
    }
    assert simple.parameters["std_err"].to_series().to_dict() == approx(
        expected_stderr, rel=2e-3
    )


def test_mixed_logit(simple_model: lx.Model):
    mixed = simple_model.copy()
    mixed.mixtures = [
        lx.mixtures.Normal(k, f"s_{k}")
        for k in ["opcost", "max_range", "ev", "hybrid", "hiperf", "medhiperf"]
    ]
    mixed.n_draws = 200
    mixed.seed = 42
    assert mixed.pvals == approx(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001]
    )
    mixed.pvals = "null"
    assert mixed.pvals == approx(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )
    assert mixed.compute_engine == "jax"
    assert mixed.n_draws == 200
    assert mixed.seed == 42
    assert len(mixed.mixtures) == 6
    assert isinstance(mixed.mixtures[0], lx.mixtures.Normal)
    assert mixed.mixtures[0].to_dict() == {
        "type": "Normal",
        "mean": "opcost",
        "std": "s_opcost",
    }

    # Check initial values are a good match
    assert mixed.loglike() == approx(-1630.34033203125)
    np.testing.assert_allclose(
        mixed.d_loglike(),
        np.array(
            [
                -2.056667e02,
                5.766666e01,
                6.666666e01,
                -2.477668e02,
                8.300000e01,
                2.914683e02,
                -7.866664e02,
                -3.433672e-02,
                -2.093602e-02,
                -2.493681e-02,
                1.647819e-02,
                9.754598e-04,
                2.504729e-01,
            ],
            dtype=np.float32,
        ),
        rtol=1e-5,
    )

    result0 = mixed.maximize_loglike(stderr=True, options={"maxiter": 2})
    assert not result0["success"]
    assert result0["nit"] == 2
    assert result0["loglike"] == approx(-1565.2197265625)
    expected_value = {
        "ev": -0.14029760159652338,
        "hiperf": 0.03933788979469556,
        "hybrid": 0.04547732430870773,
        "max_range": -0.1690165820789483,
        "medhiperf": 0.05661927740375723,
        "opcost": 0.19882802193266583,
        "price": -0.5366323560619928,
        "s_ev": -2.3423133069298892e-05,
        "s_hiperf": -1.4281714556392845e-05,
        "s_hybrid": -1.7010893622241297e-05,
        "s_max_range": 1.124076043977622e-05,
        "s_medhiperf": 6.654196362547049e-07,
        "s_opcost": 0.0001708625878473434,
    }
    assert mixed.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=5e-2
    )

    mixed.maximize_loglike(stderr=True, options={"ftol": 1e-9})
    # TEST
    assert mixed.float_dtype == np.float64
    assert mixed.most_recent_estimation_result["success"]
    assert mixed.most_recent_estimation_result["total_weight"] == approx(1484.0)
    assert mixed.most_recent_estimation_result["loglike"] == approx(-1385.494384765625)
    expected_value = {
        "ev": -3.0386047026095038,
        "hiperf": 0.1759331797395676,
        "hybrid": 0.5469828878951994,
        "max_range": 0.8913475142526257,
        "medhiperf": 0.734468888360476,
        "opcost": -0.22007540332571882,
        "price": -0.648648076713321,
        "s_ev": -2.810148184360984,
        "s_hiperf": -0.5264974590918174,
        "s_hybrid": 1.1459421445296165,
        "s_max_range": -0.12631587788430482,
        "s_medhiperf": 1.5788576158691365,
        "s_opcost": 0.5213287408633901,
    }
    assert mixed.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=5e-2
    )
    expected_stderr = {
        "ev": 0.7826659083366394,
        "hiperf": 0.12497398257255554,
        "hybrid": 0.19386492669582367,
        "max_range": 0.32467690110206604,
        "medhiperf": 0.2096310406923294,
        "opcost": 0.0717368870973587,
        "price": 0.10241813212633133,
        "s_ev": 0.7595077753067017,
        "s_hiperf": 0.8704573512077332,
        "s_hybrid": 0.6614100337028503,
        "s_max_range": 1.0017218589782715,
        "s_medhiperf": 0.6141265034675598,
        "s_opcost": 0.1859593540430069,
    }
    assert mixed.parameters["std_err"].to_series().to_dict() == approx(
        expected_stderr, rel=5e-2
    )

    # make sure we can copy the model
    mixed2 = mixed.copy()
    assert mixed2.compute_engine == "jax"
    assert mixed2.n_draws == 200
    assert mixed2.seed == 42
    assert len(mixed2.mixtures) == 6
    assert isinstance(mixed2.mixtures[0], lx.mixtures.Normal)
    assert mixed2.mixtures[0].to_dict() == {
        "type": "Normal",
        "mean": "opcost",
        "std": "s_opcost",
    }
    assert mixed.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=5e-2
    )
    assert mixed.parameters["std_err"].to_series().to_dict() == approx(
        expected_stderr, rel=5e-2
    )



from __future__ import annotations

import numpy as np
import pandas as pd
from pytest import fixture, raises

import larch as lx
from larch.examples import example_file


@fixture(scope="module")
def mtc_dataset_idca():
    ca = pd.read_csv(
        example_file("MTCwork.csv.gz"),
        index_col=("casenum", "altnum"),
        dtype={"chose": np.float32},
    )
    return lx.Dataset.dc.from_idca(ca)


def test_datarray_dc_accessor(mtc_dataset_idca):
    ivtt = mtc_dataset_idca.dc["ivtt"]
    assert ivtt.shape == (5029, 6)
    assert ivtt.dc.n_cases == 5029
    assert ivtt.dc.n_alts == 6

    dist = mtc_dataset_idca.dc["dist"]
    assert dist.shape == (5029,)
    assert dist.dc.n_cases == 5029
    with raises(ValueError):
        _ = dist.dc.n_alts


def test_dataset_dc_accessor(mtc_dataset_idca):
    assert mtc_dataset_idca.dc.n_cases == 5029
    assert mtc_dataset_idca.dc.n_alts == 6
    assert mtc_dataset_idca.dc.CASEID == "casenum"
    assert mtc_dataset_idca.dc.ALTID == "altnum"
    assert mtc_dataset_idca.dc.CASEALT is None
    assert mtc_dataset_idca.dc.CASEPTR is None
    assert mtc_dataset_idca.dc.ALTIDX is None
    assert mtc_dataset_idca.dc.GROUPID is None
    assert mtc_dataset_idca.dc.INGROUP is None


def test_dataset_dtypes():
    ca = pd.read_csv(
        example_file("MTCwork.csv.gz"),
        index_col=("casenum", "altnum"),
        dtype={"chose": np.float32},
    )

    # set some variables with different dtypes
    ca["ovtt_round"] = ca["ovtt"].round().astype(np.int8)
    ca["ivtt_round"] = ca["ivtt"].round().astype(np.uint8)
    ca["male"] = ca["femdum"] == 0

    # initialize the Dataset from the DataFrame
    ds = lx.Dataset.dc.from_idca(ca)

    # check the dtypes
    assert ds["ovtt_round"].dtype == ca["ovtt_round"].dtype
    assert ds["ivtt_round"].dtype == ca["ivtt_round"].dtype
    assert ds["male"].dtype == ca["male"].dtype
    assert ds["chose"].dtype == ca["chose"].dtype
    assert ds["ivtt"].dtype == ca["ivtt"].dtype
    assert ds["hhid"].dtype == ca["hhid"].dtype



from __future__ import annotations

import pandas as pd
import pytest
from pytest import approx

import larch as lx
from larch import P, X


@pytest.fixture
def swissmetro_data():
    raw = pd.read_csv(lx.example_file("swissmetro.csv.gz"))
    raw["SM_COST"] = raw["SM_CO"] * (raw["GA"] == 0)
    raw["TRAIN_COST"] = raw.eval("TRAIN_CO * (GA == 0)")
    raw["TRAIN_COST_SCALED"] = raw["TRAIN_COST"] / 100
    raw["TRAIN_TT_SCALED"] = raw["TRAIN_TT"] / 100

    raw["SM_COST_SCALED"] = raw["SM_COST"] / 100
    raw["SM_TT_SCALED"] = raw["SM_TT"] / 100

    raw["CAR_CO_SCALED"] = raw["CAR_CO"] / 100
    raw["CAR_TT_SCALED"] = raw["CAR_TT"] / 100
    raw["CAR_AV_SP"] = raw.eval("CAR_AV * (SP!=0)")
    raw["TRAIN_AV_SP"] = raw.eval("TRAIN_AV * (SP!=0)")

    raw["keep"] = raw.eval("PURPOSE in (1,3) and CHOICE != 0")
    data = lx.Dataset.construct.from_idco(raw).dc.query_cases("keep")
    return data


def test_panel_latent_class(swissmetro_data):
    availability_co_vars = {
        1: "TRAIN_AV_SP",
        2: "SM_AV",
        3: "CAR_AV_SP",
    }

    m1 = lx.Model(title="Model1")
    m1.availability_co_vars = availability_co_vars
    m1.choice_co_code = "CHOICE"
    m1.utility_co[1] = P("ASC_TRAIN") + X("TRAIN_COST_SCALED") * P("B_COST")
    m1.utility_co[2] = X("SM_COST_SCALED") * P("B_COST")
    m1.utility_co[3] = P("ASC_CAR") + X("CAR_CO_SCALED") * P("B_COST")

    m2 = lx.Model(title="Model2")
    m2.availability_co_vars = availability_co_vars
    m2.choice_co_code = "CHOICE"
    m2.utility_co[1] = (
        P("ASC_TRAIN")
        + X("TRAIN_TT_SCALED") * P("B_TIME")
        + X("TRAIN_COST_SCALED") * P("B_COST")
    )
    m2.utility_co[2] = X("SM_TT_SCALED") * P("B_TIME") + X("SM_COST_SCALED") * P(
        "B_COST"
    )
    m2.utility_co[3] = (
        P("ASC_CAR")
        + X("CAR_TT_SCALED") * P("B_TIME")
        + X("CAR_CO_SCALED") * P("B_COST")
    )

    m3 = lx.Model(title="Model3")
    m3.availability_co_vars = availability_co_vars
    m3.choice_co_code = "CHOICE"
    m3.utility_co[1] = X("TRAIN_COST_SCALED") * P("Z_COST")
    m3.utility_co[2] = X("SM_COST_SCALED") * P("Z_COST")
    m3.utility_co[3] = X("CAR_CO_SCALED") * P("Z_COST")

    mk = lx.Model(title="ModelC")
    mk.utility_co[102] = P("W_OTHER")
    mk.utility_co[103] = P("W_COST")
    mk.groupid = "ID"

    # print("hello")

    b = lx.LatentClass(
        mk,
        {101: m1, 102: m2, 103: m3},
        datatree=swissmetro_data.dc.set_altids([1, 2, 3]),
        groupid="ID",
    )

    # print("hello2")

    # for k, m in b._models.items():
    #     print(f"==========={k=}===========\n{m.dataset}")

    b.lock_value(Z_COST=-10000)
    # for k, m in b._models.items():
    #     print(f"~=========={k=}===========\n{m.dataset}")

    assert b.loglike() == approx(-6867.245, rel=1e-4)
    # print(f"{b.loglike()=}")
    assert b.d_loglike() == approx(
        [
            -1.104770e02,
            -1.545916e03,
            -2.189165e01,
            -9.183448e02,
            -1.658521e02,
            8.292606e01,
            -1.490116e-08,
        ],
        abs=1e-7,
        rel=1e-3,
    )
    # print(f"{b.d_loglike()=}")

    result = b.maximize_loglike(method="slsqp")
    assert result.loglike == approx(-4474.478515625)
    # print(f"{result.loglike=}")


if __name__ == "__main__":
    swissmetro_data_ = swissmetro_data()
    test_panel_latent_class(swissmetro_data_)



from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from ..exceptions import BHHHSimpleStepFailure, MissingDataError
from ..util import dictx

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    pass


class ModelDashboard:
    """
    Current state of the model.

    This dashboard is displayed on creation.
    """

    def __init__(self, throttle: float = 2, visible=True, *, tags=None):
        from ..util.display import display_head, display_nothing, display_p
        from ..util.rate_limiter import NonBlockingRateLimiter

        self.visible = visible
        if tags is not None:
            self.head, self.subhead, self.body = tags
        else:
            if visible:
                self.head = display_head("Larch Model Dashboard", level=3)
                self.subhead = display_p("LL = ...")
                self.body = display_p("...")
            else:
                self.head = display_nothing()
                self.subhead = display_nothing()
                self.body = display_nothing()
        if isinstance(throttle, NonBlockingRateLimiter):
            self.throttle_gate = throttle
        else:
            self.throttle_gate = NonBlockingRateLimiter(throttle)

    def update(self, head=None, subhead=None, body=None, force=False):
        if self.visible:
            if self.throttle_gate or force or 1:
                if head is not None:
                    self.head.update(head, force=force)
                if subhead is not None:
                    self.subhead.update(subhead, force=force)
                if body is not None:
                    self.body.update(body, force=force)


def maximize_loglike(
    model,
    method=None,
    method2=None,
    quiet=False,
    screen_update_throttle=2,
    final_screen_update=True,
    check_for_overspecification=True,
    return_tags=False,
    reuse_tags=None,
    iteration_number=0,
    iteration_number_tail="",
    options=None,
    maxiter=None,
    bhhh_start=0,
    jumpstart=0,
    jumpstart_split=5,
    return_dashboard=False,
    dashboard=None,
    prior_result=None,
    stderr=False,
    **kwargs,
) -> dictx:
    """
    Maximize the log likelihood.

    Parameters
    ----------
    model : AbstractChoiceModel
        The data for this model should previously have been
        prepared using the `load_data` method.
    method : str, optional
        The optimization method to use.  See scipy.optimize for
        most possibilities, or use 'BHHH'. Defaults to SLSQP if
        there are any constraints or finite parameter bounds,
        otherwise defaults to BHHH.
    quiet : bool, default False
        Whether to suppress the dashboard.

    Returns
    -------
    larch.util.dictx
        A dictionary of results, including final log likelihood,
        elapsed time, and other statistics.  The exact items
        included in output will vary by estimation method.

    Raises
    ------
    ValueError
        If the `dataframes` are not already loaded.

    """
    _initial_constraint_intensity = getattr(model, "constraint_intensity", None)
    _initial_constraint_sharpness = getattr(model, "constraint_sharpness", None)
    try:
        from scipy.optimize import minimize

        from ..util.timesize import Timer

        _doctest_mode_ = False
        from .numbamodel import NumbaModel

        if isinstance(model, NumbaModel):
            if (
                getattr(model, "data_as_loaded", None) is None
                and getattr(model, "datatree", None) is not None
            ):
                model.unmangle(force=True)
            if (
                getattr(model, "data_as_loaded", None) is None
                and not model.use_streaming
            ):
                raise MissingDataError("no data attached to model")

        if prior_result is not None:
            dashboard = dashboard or prior_result.get("dashboard", None)
            iteration_number = iteration_number or prior_result.get(
                "iteration_number", 0
            )

        if _doctest_mode_:
            from ..model import Model

            if type(model) is Model:
                model.unmangle()
                model._frame.sort_index(inplace=True)
                model.unmangle(True)

        if options is None:
            options = {}
        if maxiter is not None:
            options["maxiter"] = maxiter

        timer = Timer()

        if not quiet and not _doctest_mode_:
            if (
                isinstance(reuse_tags, tuple)
                and len(reuse_tags) == 3
                and dashboard is None
            ):
                dashboard = ModelDashboard(screen_update_throttle, tags=reuse_tags)
            if dashboard is not None:
                if not isinstance(dashboard, ModelDashboard):
                    raise ValueError(
                        f"dashboard must be ModelDashboard, not {type(dashboard)}"
                    )
            else:
                dashboard = ModelDashboard(screen_update_throttle)
        else:
            dashboard = ModelDashboard(visible=False)

        def callback(x, status=None):
            nonlocal iteration_number, dashboard, method
            iteration_number += 1
            if isinstance(status, dict) and "penalty" in status:
                dashboard.update(
                    f"Iteration {iteration_number:03} {iteration_number_tail}",
                    (
                        f"Currently using {method}, "
                        f"Best LL = {model._cached_loglike_best}, "
                        f"Current Total LL = {status['total_loglike']}, "
                        f"Constraint Penalty = {status['penalty']}"
                    ),
                    model.pf,
                )
            else:
                dashboard.update(
                    f"Iteration {iteration_number:03} {iteration_number_tail}",
                    f"Currently using {method}, Best LL = {model._cached_loglike_best}",
                    model.pf,
                )
            return False

        if quiet or _doctest_mode_:
            callback = None  # noqa: F811

        if bhhh_start:
            if method is None or method.lower() == "bhhh":
                method = "slsqp"

        if method is None:
            try:
                has_constraints = bool(model.constraints)
            except AttributeError:
                has_constraints = False
            if (
                has_constraints
                or np.isfinite(model.pminimum.max())
                or np.isfinite(model.pmaximum.min())
            ):
                method = "slsqp"
            else:
                method = "bhhh"

        if method2 is None and method.lower() == "bhhh":
            method2 = "slsqp"

        method_used = method
        raw_result = None

        if bhhh_start:
            try:
                _restore_method = method
                method_used = f"bhhh({bhhh_start})->{method}"
                method = f"bhhh({bhhh_start})"
                current_ll, tolerance, iter, steps_bhhh, message = model.fit_bhhh(
                    steplen=1.0,
                    momentum=5,
                    logger=None,
                    ctol=1e-4,
                    maxiter=options.get("maxiter", 100),
                    soft_maxiter=bhhh_start,
                    callback=callback,
                    minimum_steplen=0.0001,
                    maximum_steplen=1.0,
                    initial_constraint_intensity=1.0,
                    step_constraint_intensity=1.5,
                    max_constraint_intensity=1e6,
                    initial_constraint_sharpness=1.0,
                    step_constraint_sharpness=1.5,
                    max_constraint_sharpness=1e6,
                )
                raw_result = {
                    "loglike": current_ll,
                    "x": model.pvals,
                    "tolerance": tolerance,
                    "steps": steps_bhhh,
                    "message": message,
                }
            except BHHHSimpleStepFailure:
                dashboard.update(
                    f"Iteration {iteration_number:03} "
                    f"[BHHH Start Failure, Recovering] {iteration_number_tail}",
                    body=model.pf,
                    force=True,
                )
            finally:
                method = _restore_method
                model.constraint_intensity = 0.0

        if method.lower() == "bhhh":
            try:
                max_iter = options.get("maxiter", 100)
                stopping_tol = options.get("ctol", 1e-5)

                if hasattr(model, "fit_bhhh"):
                    (
                        current_ll,
                        tolerance,
                        iter_bhhh,
                        steps_bhhh,
                        message,
                    ) = model.fit_bhhh(
                        # steplen=1.0,
                        # momentum=5,
                        ctol=stopping_tol,
                        maxiter=max_iter,
                        callback=callback,
                    )
                else:
                    (
                        current_ll,
                        tolerance,
                        iter_bhhh,
                        steps_bhhh,
                        message,
                    ) = model.simple_fit_bhhh(
                        ctol=stopping_tol,
                        maxiter=max_iter,
                        callback=callback,
                        jumpstart=jumpstart,
                        jumpstart_split=jumpstart_split,
                    )
                raw_result = {
                    "loglike": current_ll,
                    "x": model.pvals,
                    "tolerance": tolerance,
                    "steps": steps_bhhh,
                    "message": message,
                }
            except (NotImplementedError, AttributeError):
                dashboard.update(
                    f"Iteration {iteration_number:03} [BHHH Not Available] {iteration_number_tail}",
                    body=model.pf,
                    force=True,
                )
                if method2 is not None:
                    method_used = f"{method2}"
                    method = method2
            except BHHHSimpleStepFailure:
                dashboard.update(
                    f"Iteration {iteration_number:03} [Exception Recovery] {iteration_number_tail}",
                    body=model.pf,
                    force=True,
                )
                if method2 is not None:
                    method_used = f"{method_used}|{method2}"
                    method = method2
            except Exception:
                dashboard.update(
                    f"Iteration {iteration_number:03} [Exception] {iteration_number_tail}",
                    body=model.pf,
                    force=True,
                )
                raise

        if method.lower() != "bhhh":
            try:
                bounds = None
                if isinstance(method, str) and method.lower() in (
                    "slsqp",
                    "l-bfgs-b",
                    "tnc",
                    "trust-constr",
                ):
                    bounds = model.pbounds
                    if np.any(
                        (model.pholdfast == 0) & np.isinf(model.pminimum)
                    ) or np.any((model.pholdfast == 0) & np.isinf(model.pmaximum)):
                        warnings.warn(  # infinite bounds #  )
                            f"{method} may not play nicely with unbounded parameters\n"
                            "if you get poor results, consider setting global bounds "
                            "with model.set_cap()",
                            stacklevel=1,
                        )

                try:
                    constraints = model._get_constraints(method)
                except Exception:
                    constraints = ()

                # args = getattr(model, "_null_slice", (0, -1, 1))
                raw_result = minimize(
                    model.logloss,
                    model.pvals,
                    # args=args,
                    method=method,
                    jac=model.d_logloss,
                    bounds=bounds,
                    callback=callback,
                    options=options,
                    constraints=constraints,
                    **kwargs,
                )
            except Exception:
                dashboard.update(
                    f"Iteration {iteration_number:03} [Exception] {iteration_number_tail}",
                    body=model.pf,
                    force=True,
                )
                raise

        if stderr:
            model.calculate_parameter_covariance()

        timer.stop()

        if (
            final_screen_update
            and not quiet
            and not _doctest_mode_
            and raw_result is not None
        ):
            converged = raw_result.get("message", "Converged")
            dashboard.update(
                f"Iteration {iteration_number:03} [{converged}] {iteration_number_tail}",
                f"Best LL = {model._cached_loglike_best}",
                model.pf,
                force=True,
            )

        if raw_result is None:
            raw_result = {}
        # if check_for_overspecification:
        # 	model.check_for_possible_overspecification()

        result = dictx()
        for k, v in raw_result.items():
            if k == "fun":
                result["logloss"] = v
            elif k == "jac":
                try:
                    result["d_logloss"] = pd.Series(-v, index=model.pnames)
                except TypeError:
                    result[k] = v
            elif k == "x":
                result["x"] = pd.Series(v, index=model.pnames)
            else:
                result[k] = v
        result["elapsed_time"] = timer.elapsed()
        result["method"] = method_used
        try:
            result["n_cases"] = model.n_cases
        except (NotImplementedError, AttributeError):
            pass
        result["iteration_number"] = iteration_number

        if "logloss" in result:
            result["loglike"] = -result["logloss"] * model.total_weight()

        if _doctest_mode_:
            result["__verbose_repr__"] = True

        model._most_recent_estimation_result = result.copy()

        if return_dashboard:
            result["dashboard"] = dashboard

        if return_tags:
            return result, dashboard.head, dashboard.subhead, dashboard.body

        return result

    except Exception:
        logger.exception("error in maximize_loglike")
        raise
    finally:
        if _initial_constraint_intensity is not None:
            model.constraint_intensity = _initial_constraint_intensity
        if _initial_constraint_sharpness is not None:
            model.constraint_sharpness = _initial_constraint_sharpness


def propose_direction(bhhh, dloglike, freedoms):
    direction = np.zeros_like(dloglike)
    # try:
    #     direction1 = np.linalg.solve(bhhh[freedoms, :][:, freedoms], dloglike[freedoms])
    # except np.linalg.LinAlgError:
    direction1 = np.linalg.lstsq(bhhh[freedoms, :][:, freedoms], dloglike[freedoms])[0]
    try:
        direction[freedoms] = direction1
    except Exception:
        print("direction", direction.shape)
        print("direction1", direction1.shape)
        print("freedoms", freedoms.shape)
        raise
    return direction


def fit_bhhh(
    model,
    steplen=1.0,
    momentum=5,
    printer=None,
    ctol=1e-4,
    maxiter=100,
    callback=None,
    jumpstart=0,
    jumpstart_split=5,
    minimum_steplen=0.0001,
    maximum_steplen=1.0,
    initial_constraint_intensity=None,
    step_constraint_intensity=1.5,
    initial_constraint_sharpness=None,
    step_constraint_sharpness=1.2,
):
    """
    Make a series of steps using the BHHH algorithm.

    Parameters
    ----------
    steplen: float
    printer: callable

    Returns
    -------
    loglike, convergence_tolerance, n_iters, steps
    """
    current_pvals = model.pvals.copy()
    iter = 0
    steps = []

    if initial_constraint_intensity is not None:
        model.constraint_intensity = initial_constraint_intensity
    if initial_constraint_sharpness is not None:
        model.constraint_sharpness = initial_constraint_sharpness

    # if jumpstart:
    #     model.jumpstart_bhhh(jumpstart=jumpstart, jumpstart_split=jumpstart_split)
    #     iter += jumpstart

    current_ll, current_dll, current_bhhh = model._loglike2_bhhh_tuple()

    def find_direction(current_dll, current_bhhh):
        freedoms = model.pholdfast == 0
        direction = propose_direction(current_bhhh, np.asarray(current_dll), freedoms)
        tolerance = np.dot(direction, current_dll)
        return direction, tolerance

    direction, tolerance = find_direction(current_dll, current_bhhh)

    while abs(tolerance) > ctol and iter < maxiter:
        iter += 1
        if steps:
            steplen = min(
                2.0 * sum(steps[-momentum:]) / len(steps[-momentum:]), maximum_steplen
            )
        while True:
            model.pvals = current_pvals + direction * steplen
            proposed_ll, proposed_dll, proposed_bhhh = model._loglike2_bhhh_tuple()
            if proposed_ll > current_ll:
                break
            steplen *= 0.5
            if steplen < minimum_steplen:
                break
        if proposed_ll <= current_ll:
            model.set_values(current_pvals)
            raise BHHHSimpleStepFailure(
                f"simple step bhhh failed\ndirection = {str(direction)}"
            )
        if printer is not None:
            printer(f"simple step bhhh {steplen} to gain {proposed_ll - current_ll}")
        steps.append(steplen)

        current_ll, current_dll, current_bhhh = proposed_ll, proposed_dll, proposed_bhhh
        current_pvals = model.pvals.copy()
        if callback is not None:
            callback(current_pvals)

        model.constraint_intensity *= step_constraint_intensity
        model.constraint_sharpness *= step_constraint_sharpness
        direction, tolerance = find_direction(current_dll, current_bhhh)

    if abs(tolerance) <= ctol:
        message = "Optimization terminated successfully"
    else:
        message = f"Optimization terminated after {iter} iterations"

    return current_ll, tolerance, iter, np.asarray(steps), message



from __future__ import annotations

from numbers import Number

from .linear import ParameterRef, _what_is

_precedence = {
    "lambda": 0,
    "if": 1,
    "or": 2,
    "and": 3,
    "not": 4,
    "compare": 5,
    "|": 6,
    "^": 7,
    "&": 8,
    ">>": 9,
    "+": 10,
    "*": 11,
    "~": 12,
    "**": 13,
}


def _parenthize_repr(obj, local_precedence):
    obj_precedence = getattr(obj, "_precedence", 0)
    if obj_precedence < local_precedence:
        return f"({repr(obj)})"
    else:
        return repr(obj)


class _ParameterOp:
    def set_fmt(self, formatting):
        self._formatting = formatting
        return self

    def value(self, *args):
        raise NotImplementedError("abstract base class")

    def string(self, m):
        """
        Get the value of the parameter math from a given model, as a formatted string.

        Parameters
        ----------
        m : Model
                The model from which to extract a parameter value.

        Returns
        -------
        str
        """
        if self._formatting is None:
            return f"{self.value(m):.3g}"
        else:
            return self._formatting.format(self.value(m))

    def __add__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterAdd(self, other)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterAdd(self, other.as_pmath())
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __radd__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterAdd(other, self)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterAdd(other.as_pmath(), self)
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(other)} + {_what_is(self)}")

    def __sub__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterSubtract(self, other)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterSubtract(self, other.as_pmath())
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(self)} - {_what_is(other)}")

    def __rsub__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterSubtract(other, self)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterSubtract(other.as_pmath(), self)
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(other)} - {_what_is(self)}")

    def __mul__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterMultiply(self, other)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterMultiply(self, other.as_pmath())
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(self)} * {_what_is(other)}")

    def __rmul__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterMultiply(other, self)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterMultiply(other.as_pmath(), self)
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(other)} * {_what_is(self)}")

    def __truediv__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterDivide(self, other)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterDivide(self, other.as_pmath())
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(self)} / {_what_is(other)}")

    def __rtruediv__(self, other):
        if isinstance(other, (ParameterRef, _ParameterOp, Number)):
            return ParameterDivide(other, self)
        if hasattr(other, "as_pmath"):
            try:
                return ParameterDivide(other.as_pmath(), self)
            except Exception:
                pass
        raise NotImplementedError(f"{_what_is(other)} / {_what_is(self)}")


ParameterOp = _ParameterOp


class _ParameterUnaryOp(_ParameterOp):
    _op = "??"
    _precedence = 0

    def __init__(self, operand):
        self._operand = operand
        self._formatting = getattr(operand, "_formatting", None)

    def valid(self, m):
        return self._operand.valid(m)

    def __repr__(self):
        return f"{self._op}{_parenthize_repr(self._operand, self._precedence)}"


class _ParameterBinaryOp(_ParameterOp):
    _op = "??"
    _precedence = 0

    def __init__(self, left, right):
        self._left = left
        self._right = right
        self._formatting = getattr(left, "_formatting", None) or getattr(
            right, "_formatting", None
        )

    def valid(self, m):
        if isinstance(self._left, (ParameterRef, _ParameterOp)):
            x = self._left.valid(m)
        else:
            x = True
        if isinstance(self._right, (ParameterRef, _ParameterOp)):
            x &= self._right.valid(m)
        return x

    def __repr__(self):
        return (
            f"{_parenthize_repr(self._left, self._precedence)} "
            f"{self._op} "
            f"{_parenthize_repr(self._right, self._precedence)}"
        )

    def __str__(self):
        return (
            f"{_parenthize_repr(self._left, self._precedence)} "
            f"{self._op} "
            f"{_parenthize_repr(self._right, self._precedence)}"
        )


class ParameterAdd(_ParameterBinaryOp):
    _op = "+"
    _precedence = 10

    def value(self, *args):
        if isinstance(self._left, (ParameterRef, _ParameterOp)):
            x = self._left.value(*args)
        else:
            x = self._left
        if isinstance(self._right, (ParameterRef, _ParameterOp)):
            x += self._right.value(*args)
        else:
            x += self._right
        return x


class ParameterMultiply(_ParameterBinaryOp):
    _op = "*"
    _precedence = 11

    def value(self, *args):
        if isinstance(self._left, (ParameterRef, _ParameterOp)):
            x = self._left.value(*args)
        else:
            x = self._left
        if isinstance(self._right, (ParameterRef, _ParameterOp)):
            x *= self._right.value(*args)
        else:
            x *= self._right
        return x


class ParameterSubtract(_ParameterBinaryOp):
    _op = "-"
    _precedence = 10

    def value(self, *args):
        if isinstance(self._left, (ParameterRef, _ParameterOp)):
            x = self._left.value(*args)
        else:
            x = self._left
        if isinstance(self._right, (ParameterRef, _ParameterOp)):
            x -= self._right.value(*args)
        else:
            x -= self._right
        return x


class ParameterDivide(_ParameterBinaryOp):
    _op = "/"
    _precedence = 11

    def value(self, *args):
        if isinstance(self._left, (ParameterRef, _ParameterOp)):
            x = self._left.value(*args)
        else:
            x = self._left
        if isinstance(self._right, (ParameterRef, _ParameterOp)):
            x /= self._right.value(*args)
        else:
            x /= self._right
        return x


class ParameterNoop(_ParameterUnaryOp):
    _op = ""
    _precedence = 98

    def value(self, *args):
        return self._operand.value(*args)


class ParameterNegate(_ParameterUnaryOp):
    _op = "-"
    _precedence = 12

    def value(self, *args):
        return -self._operand.value(*args)
